#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""


"""

from collections.abc import Iterable
import numpy as np
from scipy.ndimage.filters import gaussian_filter
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.colors import LogNorm
from mpl_toolkits.axes_grid1 import make_axes_locatable
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import seislib.colormaps as scm



def add_inset_axis(axis, rect, polar=False):
    """ Adds an axis inset
    
    Parameters
    ----------
    axis : matplotlib AxesSubplot
        Instance of `matplotlib.pyplot.subplot`
        
    rect : list or tuple of shape (4,)
        [bottom, left, width, height] of the inset. Bottom and left 
        coordinates are expressed with respect to `axis`
        
    polar : bool
        If `True`, a polar plot is created. Default is `False`
        
    
    Returns
    -------
    Inset axis
    """
    def axis_to_fig(axis):
        fig = axis.figure
        def transform(coord):
            return fig.transFigure.inverted().transform(
                axis.transAxes.transform(coord))
        return transform
    
    fig = axis.figure
    left, bottom, width, height = rect
    trans = axis_to_fig(axis)
    figleft, figbottom = trans((left, bottom))
    figwidth, figheight = trans([width,height]) - trans([0,0])
    return fig.add_axes([figleft, figbottom, figwidth, figheight], polar=polar)


def lower_threshold_projection(projection, thresh=1e3, **kwargs):
    """ 
    Effective work around to get a higher-resolution curvature of the 
    great-circle paths to be plotted in a given projection. This is useful 
    when plotting the great-circle paths in a relatively small region.
    
    Parameters
    ----------
    projection : class
        Should be one of the cartopy projection classes, e.g., 
        `cartopy.crs.Mercator`
        
    thresh : float
        Smaller values achieve higher resolutions. Default is 1e3
        
    Returns
    ------- 
    Instance of the input (`projection`) class
    
    
    Examples
    --------
    >>> proj = lower_threshold_projection(cartopy.crs.Mercator, 
    ...                                   thresh=1e3)
    
    Note that the `cartopy.crs.Mercator` was not initialized (i.e., 
    there are no brackets after the word `Mercator`)
    """
    class LowerThresholdProjection(projection):
        
        @property
        def threshold(self):
            return thresh
        
    return LowerThresholdProjection(**kwargs)

    
def add_earth_features_GSHHS(ax, scale='auto', oceans_color='aqua', 
                             lands_color='coral', edgecolor='face'):
    """ 
    Adds natural features to a `cartopy.mpl.geoaxes.GeoAxesSubplot`, fetching data
    from the `GSHHS dataset <https://www.ngdc.noaa.gov/mgg/shorelines/gshhs.html>`_
    
    Parameters
    ----------
    ax : cartopy.mpl.geoaxes.GeoAxesSubplot
    
    scale : {'auto', 'coarse', 'low', 'intermediate', 'high', 'full'}
        Scale of the dataset, passed to `cartopy.feature.GSHHSFeature`. 
        Default is 'auto'
        
    oceans_color, lands_color : str
        Color of oceans and continents in the map. They should be valid 
        `matplotlib` colors or be part of `cartopy.cfeature.COLORS`. Defaults 
        are 'aqua' and 'coral'
        
    edgecolor : str
        Color of the boundaries between, e.g., lakes and land. Passed to 
        `cartopy.feature.GSHHSFeature`. Default is 'face' (the boundaries will
        have the same color as the continent)
    """
    if oceans_color == 'water':
        oceans_color = cfeature.COLORS[oceans_color]
    if lands_color == 'land':
        lands_color = cfeature.COLORS[lands_color]
    colors = [lands_color, oceans_color] * 2
    ax.background_patch.set_facecolor(oceans_color)
    for level, color in zip(range(1, 5), colors):
        feature = cfeature.GSHHSFeature(levels=[level], 
                                        scale=scale,
                                        edgecolor=edgecolor, 
                                        facecolor=color)
        ax.add_feature(feature)



def add_earth_features(ax, scale='110m', oceans_color='aqua', 
                       lands_color='coral', edgecolor='k', lands_lw=0.5,
                       oceans_lw=0.5, lakes_lw=0.5):
    """ 
    Adds natural features to a `cartopy.mpl.geoaxes.GeoAxesSubplot`, fetching
    data from the `Natural Earth dataset <http://www.naturalearthdata.com/>`_
    
    Parameters
    ----------
    ax : cartopy.mpl.geoaxes.GeoAxesSubplot
    
    scale : {'10m', '50m', '110m'}
        Resolution of the Earth features displayed in the figure. Passed to
        `cartopy.feature.NaturalEarthFeature`. Default is '110m'
        
    oceans_color, lands_color : str
        Color of oceans and continents in the map. They should be valid 
        `matplotlib` colors or be part of `cartopy.cfeature.COLORS`. Defaults 
        are 'aqua' and 'coral'
        
    edgecolor : str
        Color of the boundaries between, e.g., lakes and land. Passed to 
        `cartopy.feature.NaturalEarthFeature`. Default is 'k' (black)
        
    lands_lw, oceans_lw, lakes_lw : float
        Linewidths for lands, oceans, and lakes. Arguments passed to
        `GeoAxesSubplot.add_feature`
    """
    if oceans_color == 'water':
        oceans_color = cfeature.COLORS[oceans_color]
    if lands_color == 'land':
        lands_color = cfeature.COLORS[lands_color]
    land = cfeature.NaturalEarthFeature('physical', 
                                        'land', 
                                        scale=scale, 
                                        edgecolor=edgecolor, 
                                        facecolor=lands_color)
    ocean = cfeature.NaturalEarthFeature('physical', 
                                         'ocean',
                                         scale=scale, 
                                         edgecolor=edgecolor, 
                                         facecolor=oceans_color)
    lakes = cfeature.NaturalEarthFeature('physical', 
                                         'lakes',
                                         scale=scale, 
                                         edgecolor=edgecolor, 
                                         facecolor=oceans_color)
    ax.add_feature(land, linewidth=lands_lw)
    ax.add_feature(ocean, linewidth=oceans_lw)
    ax.add_feature(lakes, linewidth=lakes_lw)
    

def make_colorbar(ax, mappable, size='5%', pad='3%', **kwargs):
    """ Prepares and attaches a colorbar to the GeoAxesSubplot
    
    Parameters
    ----------
    ax : cartopy.mpl.geoaxes.GeoAxesSubplot
    
    mappable : matplotlib.cm.ScalarMappable
    
    size : str
        Width of the colorbar, default is '5%'
    
    pad : str
        Space between the colorbar and ax, default is '3%'
    
    **kwargs
        Additional keyword arguments passed to
        `matplotlib.pyplot.colorbar`
    
    
    Returns
    -------
    matplotlib.colorbar.Colorbar
    """
    divider = make_axes_locatable(ax)
    orientation = kwargs.pop('orientation', 'horizontal')
    if orientation == 'vertical':
        loc = 'right'
    elif orientation == 'horizontal':
        loc = 'bottom'  
    cax = divider.append_axes(loc, size, pad=pad, axes_class=mpl.pyplot.Axes)
    cb = ax.get_figure().colorbar(mappable, 
                                  cax=cax, 
                                  orientation=orientation,
                                  **kwargs)
    return cb
    

def colormesh(mesh, c, ax, **kwargs):
    """
    Adaptation of `matplotlib.pyplot.pcolormesh` to the (adaptive) equal-area 
    grid
    
    Parameters
    ----------
    mesh : ndarray (n, 4)
        Equal area grid, consisting of `n` blocks described by lat1, lat2, 
        lon1, lon2
        
    c : list of ndarray (n,)
        Values to plot in each grid cell
        
    ax : cartopy.mpl.geoaxes.GeoAxesSubplot
        Earth image on which `c` will be plotted
        
    **kwargs
        Additional keyword arguments passed to `ax.pcolormesh`
    
    
    Returns
    -------
        img : Instance of `matplotlib.collections.QuadMesh`
    """
    def plot_band(lat1, lat2, lons, v, **kwargs):
        lat_mesh, lon_mesh = np.meshgrid([lat1, lat2], lons)
        img = ax.pcolormesh(lon_mesh, lat_mesh, np.reshape(v, (-1, 1)), 
                            transform=ccrs.PlateCarree(), **kwargs)
        return img
    
    def update_kwargs(c, **kwargs):
        updated = {i: j for i, j in kwargs.items()}
        norm = updated.get('norm')
        if norm is None:
            if updated.get('vmin') is None:
                updated['vmin'] = np.nanmin(c)
            if updated.get('vmax') is None:
                updated['vmax'] = np.nanmax(c)
        else:
            if norm.vmin is None:
                norm.vmin = updated.pop('vmin', np.nanmin(c))
            if norm.vmax is None:
                norm.vmax = updated.pop('vmax', np.nanmax(c))
        return updated

    kwargs = update_kwargs(c, **kwargs)
    
    lat1, lat2 = mesh[0, :2]
    lons = [mesh[0, 2]]
    v = []
    for i, (newlat1, newlat2, lon1, lon2) in enumerate(mesh):
        if newlat1!=lat1 or newlat2!=lat2 or lon1!=lons[-1]:
            img = plot_band(lat1, lat2, lons, v, **kwargs)
            lat1, lat2 = newlat1, newlat2
            lons = [lon1, lon2]
            v = [c[i]]
        else:
            lons.append(lon2)
            v.append(c[i])
    if v:
        img = plot_band(lat1, lat2, lons, v, **kwargs)
    return img


def contourf(mesh, c, ax, smoothing=None, **kwargs):
    """
    Adaptation of matplotlib.pyplot.contourf to the (adaptive) equal area 
    grid.
    
    Parameters
    ----------
    mesh : ndarray (n, 4)
        Equal area grid, consisting of n pixels described by lat1, lat2, 
        lon1, lon2
        
    c : list of ndarray (n,)
        Values to plot in each grid cell
        
    ax : cartopy.mpl.geoaxes.GeoAxesSubplot
        Earth image on which `c` will be plotted
        
    smoothing : float, optional
        Value passed to `scipy.ndimage.filters.gaussian_filter` and used to
        obtain a smooth representation of `c` (default is `None`)
        
    **kwargs
        Additional keyword arguments passed to `ax.contourf`
    
    Returns
    -------
        img : Instance of `matplotlib.contour.QuadContourSet`
    """
    def update_kwargs(c, **kwargs):
        if kwargs.get('norm') is None:
            return kwargs
        
        updated = {i: j for i, j in kwargs.items()}
        norm = updated.get('norm')       
        if norm.vmin is None:
            norm.vmin = updated.pop('vmin', np.nanmin(c))
        if norm.vmax is None:
            norm.vmax = updated.pop('vmax', np.nanmax(c))
            
        if type(norm) == type(LogNorm()):
            levels = updated.get('levels', 50)
            lev_exp = np.linspace(np.log10(norm.vmin), 
                                  np.log10(norm.vmax), 
                                  levels)
            updated['levels'] = np.power(10, lev_exp)
            
        return updated

    kwargs = update_kwargs(c, **kwargs)
    if smoothing is not None:
        c = gaussian_filter(c, sigma=smoothing)
    lats = np.mean(mesh[:, :2], axis=1)
    lons = np.mean(mesh[:, 2:], axis=1)
    img = ax.tricontourf(lons, lats, c, transform=ccrs.PlateCarree(), **kwargs)
    return img
     

def contour(mesh, c, ax, smoothing=None, **kwargs):
    """
    Adaptation of `matplotlib.pyplot.contour` to the (adaptive) equal area 
    grid
    
    Parameters
    ----------
    mesh : ndarray (n, 4)
        Equal area grid, consisting of n pixels described by lat1, lat2, 
        lon1, lon2
        
    c : list of ndarray (n,)
        Values to plot in each grid cell
        
    ax : cartopy.mpl.geoaxes.GeoAxesSubplot
        Earth image on which `c` will be plotted
        
    smoothing : float, optional
        Value passed to `scipy.ndimage.filters.gaussian_filter` and used to
        obtain a smooth representation of `c` (default is `None`)
        
    **kwargs
        Additional keyword arguments passed to ax.contour
    
    
    Returns
    -------
        img : Instance of `matplotlib.contour.QuadContourSet`
    """
    
    def update_kwargs(c, **kwargs):
        if kwargs.get('norm') is None:
            return kwargs
        
        updated = {i: j for i, j in kwargs.items()}
        norm = updated.get('norm')       
        if norm.vmin is None:
            norm.vmin = updated.pop('vmin', np.nanmin(c))
        if norm.vmax is None:
            norm.vmax = updated.pop('vmax', np.nanmax(c))
            
        if type(norm) == type(LogNorm()):
            levels = updated.get('levels', 50)
            lev_exp = np.linspace(np.log10(norm.vmin), 
                                  np.log10(norm.vmax), 
                                  levels)
            updated['levels'] = np.power(10, lev_exp)
            
        return updated

    kwargs = update_kwargs(c, **kwargs)
    if smoothing is not None:
        c = gaussian_filter(c, sigma=smoothing)
    lats = np.mean(mesh[:, :2], axis=1)
    lons = np.mean(mesh[:, 2:], axis=1)
    img = ax.tricontour(lons, lats, c, transform=ccrs.PlateCarree(), **kwargs)
    return img


def plot_stations(stations, ax=None, show=True, oceans_color='water', 
                  lands_color='land', edgecolor='k', projection='Mercator',
                  resolution='110m', color_by_network=True, legend_dict={}, 
                  **kwargs):
    """ Creates a maps of seismic receivers
    
    Parameters
    ----------
    stations : dict
        Dictionary object containing stations information. This should 
        structured so that each key corresponds to a station code 
        ($network_code.$station_code) and each value is a tuple containing 
        latitude and longitude of the station. For example::
            
            { net1.sta1 : (lat1, lon1),
              net1.sta2 : (lat2, lon2),
              net2.sta3 : (lat3, lon3)
              }
    
    ax : cartopy.mpl.geoaxes.GeoAxesSubplot
        If not `None`, the receivers are plotted on the `GeoAxesSubplot` instance. 
        Otherwise, a new figure and `GeoAxesSubplot` instance is created
        
    show : bool
        If `True`, the plot is shown. Otherwise, a `GeoAxesSubplot` instance is
        returned. Default is `True`
        
    oceans_color, lands_color : str
        Color of oceans and lands. The arguments are ignored if ax is not
        None. Otherwise, they are passed to `cartopy.feature.NaturalEarthFeature` 
        (to the argument 'facecolor'). Defaults are 'water' and 'land'
        
    edgecolor : str
        Color of the boundaries between, e.g., lakes and land. The argument 
        is ignored if ax is not None. Otherwise, it is passed to 
        cartopy.feature.NaturalEarthFeature (to the argument 'edgecolor'). 
        Default is 'k' (black)
        
    projection : str
        Name of the geographic projection used to create the `GeoAxesSubplot`.
        (Visit the `cartopy website 
        <https://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html>`_ 
        for a list of valid projection names.) If ax is not None, `projection` 
        is ignored. Default is 'Mercator'
    
    resolution : {'10m', '50m', '110m'}
        Resolution of the Earth features displayed in the figure. Passed to
        `cartopy.feature.NaturalEarthFeature`. Default is '110m'
    
    color_by_network : bool
        If `True`, each seismic network will have a different color in the
        resulting map, and a legend will be displayed. Otherwise, all
        stations will have the same color. Default is `True`
    
    legend_dict : dict
        Keyword arguments passed to `matplotlib.pyplot.legend`
        
    kwargs : 
        Additional keyword arguments passed to `matplotlib.pyplot.scatter`
        
        
    Returns
    -------
    If `show` is True
        None

    Otherwise 
        `ax`, i.e. the `GeoAxesSubplot`
    """
    def get_coords(stations, color_by_network):
        codes, coords = zip(*[(k, v) for k, v in stations.items() \
                              if '_' not in k])
        if not color_by_network:
            yield np.array(coords), None
        else:
            networks = sorted(set(code.split('.')[0] for code in codes))
            for net in networks:
                idx = [i for i, code in enumerate(codes) if code.startswith(net)]
                yield np.array(coords)[idx], net
    
    def get_map_boundaries(stations):
        coords = np.array([i for i in stations.values()])
        latmin, latmax = np.min(coords[:,0]), np.max(coords[:,0])
        lonmin, lonmax = np.min(coords[:,1]), np.max(coords[:,1])
        dlat = (latmax - latmin) * 0.03
        dlon = (lonmax - lonmin) * 0.03
        lonmin = lonmin-dlon if lonmin-dlon > -180 else lonmin
        lonmax = lonmax+dlon if lonmax+dlon < 180 else lonmax
        latmin = latmin-dlat if latmin-dlat > -90 else latmin
        latmax = latmax+dlat if latmax+dlat < 90 else latmax
        return (lonmin, lonmax, latmin, latmax)
    
    
    transform = ccrs.PlateCarree()
    if ax is None:
        projection = eval('ccrs.%s()'%projection)
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection=projection)
        map_boundaries = get_map_boundaries(stations)
        add_earth_features(ax, 
                           scale=resolution,
                           oceans_color=oceans_color, 
                           edgecolor=edgecolor,
                           lands_color=lands_color)
        ax.set_extent(map_boundaries, transform)
    
    marker = kwargs.pop('marker', '^')
    zorder = kwargs.pop('zorder', 100)
    s = kwargs.pop('s', 100)
    for coords, net in get_coords(stations, color_by_network):
        label = kwargs.pop('label', net)
        ax.scatter(*coords.T[::-1], marker=marker, transform=transform, 
                   label=label, zorder=zorder, s=s, **kwargs) 
    
    if color_by_network:
        ax.legend(**legend_dict)
    if show:
        plt.show()
    else:
        return ax


def plot_events(lat, lon, mag=None, ax=None, show=True, oceans_color='water', 
                lands_color='land', edgecolor='k', projection='Mercator',
                resolution='110m', min_size=5, max_size=200, legend_markers=4, 
                legend_dict={}, **kwargs):
    """ Creates a map of epicenters
    
    Parameters
    ----------
    lat, lon : ndarray of shape (n,)
        Latitude and longitude of the epicenters
        
    mag : ndarray of shape(n,), optional
        If not given, the size of each on the map will be constant
    
    ax : cartopy.mpl.geoaxes.GeoAxesSubplot
        If not `None`, the receivers are plotted on the `GeoAxesSubplot` 
        instance. Otherwise, a new figure and `GeoAxesSubplot` instance is 
        created
        
    show : bool
        If `True`, the plot is shown. Otherwise, a `GeoAxesSubplot` instance is
        returned. Default is `True`
        
    oceans_color, lands_color : str
        Color of oceans and lands. The arguments are ignored if ax is not
        None. Otherwise, they are passed to `cartopy.feature.NaturalEarthFeature` 
        (to the argument 'facecolor'). Defaults are 'water' and 'land'
        
    edgecolor : str
        Color of the boundaries between, e.g., lakes and land. The argument 
        is ignored if ax is not None. Otherwise, it is passed to 
        cartopy.feature.NaturalEarthFeature (to the argument 'edgecolor'). 
        Default is 'k' (black)
        
    projection : str
        Name of the geographic projection used to create the `GeoAxesSubplot`.
        (Visit the `cartopy website 
        <https://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html>`_ 
        for a list of valid projection names.) If ax is not None, `projection` 
        is ignored. Default is 'Mercator'
    
    resolution : {'10m', '50m', '110m'}
        Resolution of the Earth features displayed in the figure. Passed to
        `cartopy.feature.NaturalEarthFeature`. Default is '110m'
    
    min_size, max_size : float, optional
        Minimum and maximum size of the epicenters on the map. These are used
        to interpolate all magnitudes associated with each event, so as to
        scale them appropriately on the map. (The final "sizes" are passed to
        the argument `s` of `matplotlib.pyplot.scatter`.) If `mag` is `None`,
        these params are ignored
        
    legend_markers : int
        Number of markers displayed in the legend. Ignored if `s` (size of the
        markers in `matplotlib.pyplot.scatter`) is passed. Only considered if
        `mag` is not `None`
            
    legend_dict : dict
        Keyword arguments passed to `matplotlib.pyplot.legend`
        
    **kwargs 
        Additional keyword arguments passed to `matplotlib.pyplot.scatter`
        
        
    Returns
    -------
    If `show` is True
        None

    Otherwise 
        `ax`, i.e. the `GeoAxesSubplot`
    """
    def get_map_boundaries(lat, lon):
        latmin, latmax = np.min(lat), np.max(lat)
        lonmin, lonmax = np.min(lon), np.max(lon)
        dlat = (latmax - latmin) * 0.03
        dlon = (lonmax - lonmin) * 0.03
        lonmin = lonmin-dlon if lonmin-dlon > -180 else lonmin
        lonmax = lonmax+dlon if lonmax+dlon < 180 else lonmax
        latmin = latmin-dlat if latmin-dlat > -90 else latmin
        latmax = latmax+dlat if latmax+dlat < 90 else latmax
        return (lonmin, lonmax, latmin, latmax)
    
    def get_markers_size(mag, kwargs):
        if mag is None:
            return kwargs.pop('s', None)
        x = np.linspace(min(mag), max(mag))
        y = np.geomspace(min_size, max_size)
        return np.interp(mag, x, y)
    
    def get_rounded_magnitudes(mag):
        magmin, magmax = min(mag), max(mag)
        return np.round(np.geomspace(magmin, magmax, legend_markers), 1)
    
    
    transform = ccrs.PlateCarree()
    if ax is None:
        projection = eval('ccrs.%s()'%projection)
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection=projection)
        map_boundaries = get_map_boundaries(lat, lon)
        ax.set_extent(map_boundaries, transform)
        add_earth_features(ax, 
                           scale=resolution,
                           oceans_color=oceans_color, 
                           edgecolor=edgecolor,
                           lands_color=lands_color)
        
    marker = kwargs.pop('marker', '*')
    zorder = kwargs.pop('zorder', 100)
    color = kwargs.pop('color', kwargs.pop('c', 'r'))
    s = kwargs.pop('s', None)
    if s is None:
        s = get_markers_size(mag, kwargs)
    if isinstance(s, Iterable):
        mags_legend = get_rounded_magnitudes(mag)
        idx = [np.argmin(np.abs(mag - mag_i)) for mag_i in mags_legend]
        idx_all = np.setdiff1d(range(len(mag)), idx)
        ax.scatter(lon[idx_all], lat[idx_all], c=color, marker=marker, 
                   transform=transform, s=s[idx_all], zorder=100, **kwargs) 
        for i, mag_legend in zip(idx, mags_legend):
            ax.scatter(lon[i], lat[i], c=color, marker=marker, transform=transform, 
                       s=s[i], label=mag_legend, zorder=100, **kwargs) 
        ax.legend(**legend_dict)
    else:
        ax.scatter(lon, lat, c=color, marker=marker, transform=transform, s=s, 
                   zorder=zorder, **kwargs) 
    if show:
        plt.show()
    else:
        return ax
    

def plot_rays(data_coords, ax=None, show=True, stations_color='r', 
              paths_color='k', oceans_color='water', lands_color='land', 
              edgecolor='k', stations_alpha=None, paths_alpha=0.3, 
              projection='Mercator', resolution='110m', map_boundaries=None, 
              bound_map=True, paths_width=0.2, **kwargs):
    """ 
    Utility function to display the great-circle paths associated with pairs
    of data coordinates
    
    Parameters
    ----------      
    data_coords : ndarray of shape (n, 4)
        Lat1, lon1, lat2, lon2 of the great-circle paths to be plotted
      
    ax : cartopy.mpl.geoaxes.GeoAxesSubplot
        If not `None`, `c` is plotted on the `GeoAxesSubplot` instance.
        Otherwise, a new figure and `GeoAxesSubplot` instance is created
        
    show : bool
        If `True` (default), the map will be showed once generated. Otherwise
        a `GeoAxesSubplot` instance is returned
        
    stations_color, paths_color : str
        Color of the receivers and of the great-circle paths (see matplotlib 
        documentation for valid color names). Defaults are 'r' (red) and
        'k' (black)
                
    oceans_color, lands_color : str
        Color of oceans and lands. The arguments are ignored if ax is not
        None. Otherwise, they are passed to `cartopy.feature.NaturalEarthFeature` 
        (to the argument 'facecolor'). Defaults are 'water' and 'land'

    edgecolor : str
        Color of the boundaries between, e.g., lakes and land. The argument 
        is ignored if ax is not None. Otherwise, it is passed to 
        cartopy.feature.NaturalEarthFeature (to the argument 'edgecolor'). 
        Default is 'k' (black)

    stations_alpha, paths_alpha : float, optional
        Transparency of the stations and of the great-circle paths. Defaults
        are `None` and 0.3
        
    paths_width : float
        Linewidth of the great-circle paths
                
    projection : str
        Name of the geographic projection used to create the `GeoAxesSubplot`.
        (Visit the `cartopy website 
        <https://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html>`_ 
        for a list of valid projection names.) If ax is not None, `projection` 
        is ignored. Default is 'Mercator'
    
    resolution : {'10m', '50m', '110m'}
        Resolution of the Earth features displayed in the figure. Passed to
        `cartopy.feature.NaturalEarthFeature`. Valid arguments are '110m',
        '50m', '10m'. Default is '110m'. Ignored if `ax` is not `None`.
    
    map_boundaries : list or tuple of floats, shape (4,), optional
        Lonmin, lonmax, latmin, latmax (in degrees) defining the extent of
        the map
        
    bound_map : bool
        If `True`, the map boundaries will be automatically determined.
        Ignored if map_boundaries is not `None`
                    

    Returns
    -------
    `None` if `show` is `True`. Otherwise a `GeoAxesSubplot` instance
    """
    def get_map_boundaries(data):
        lats = np.concatenate((data[:,0], data[:, 2]))
        lons = np.concatenate((data[:,1], data[:, 3]))
        latmin, latmax = np.nanmin(lats), np.nanmax(lats)
        lonmin, lonmax = np.nanmin(lons), np.nanmax(lons)
        dlon = (lonmax - lonmin) * 0.03
        dlat = (latmax - latmin) * 0.03
        lonmin = lonmin-dlon if lonmin-dlon > -180 else lonmin
        lonmax = lonmax+dlon if lonmax+dlon < 180 else lonmax
        latmin = latmin-dlat if latmin-dlat > -90 else latmin
        latmax = latmax+dlat if latmax+dlat < 90 else latmax
        return (lonmin, lonmax, latmin, latmax)        
    
    if ax is None:
        projection = lower_threshold_projection(eval('ccrs.%s'%projection))
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection=projection)
        add_earth_features(ax, 
                           scale=resolution,
                           oceans_color=oceans_color, 
                           edgecolor=edgecolor,
                           lands_color=lands_color)
        
    transform = ccrs.PlateCarree()
    marker = kwargs.pop('marker', '^')
    
    stations = set()
    for lat1, lon1, lat2, lon2 in data_coords:
        ax.plot([lon1, lon2], 
                [lat1, lat2], 
                color=paths_color, 
                alpha=paths_alpha, 
                lw=paths_width, 
                transform=ccrs.Geodetic())
        if (lat1, lon1) not in stations:
            ax.plot(lon1, 
                    lat1, 
                    marker=marker, 
                    c=stations_color, 
                    transform=transform, 
                    alpha=stations_alpha, 
                    **kwargs) 
            stations.add((lat1, lon1))
        if (lat2, lon2) not in stations:
            ax.plot(lon2, 
                    lat2, 
                    marker=marker, 
                    c=stations_color, 
                    transform=transform, 
                    alpha=stations_alpha, 
                    **kwargs) 
            stations.add((lat2, lon2))
            
    if map_boundaries is None and bound_map:
        map_boundaries = get_map_boundaries(data_coords)
    if map_boundaries is not None:
        ax.set_extent(map_boundaries, transform)  

    if show:
        plt.show()
    else:
        return ax    


def plot_colored_rays(data_coords, c, ax=None, show=True, cmap=scm.roma,
                      vmin=None, vmax=None, stations_color='k', 
                      oceans_color='lightgrey', lands_color='w', edgecolor='k', 
                      stations_alpha=None, paths_alpha=None, resolution='110m',
                      projection='Mercator', map_boundaries=None, bound_map=True, 
                      paths_width=1, colorbar=True, cbar_dict={}, **kwargs):
    """ 
    Utility function to display the great-circle paths associated with pairs
    of data coordinates, colored according to their respective measurements
    
    Parameters
    ----------      
    data_coords : ndarray of shape (n, 4)
        Lat1, lon1, lat2, lon2 of the great-circle paths to be plotted
        
    c : ndarray of shape (n)
        Value associated with each pair of coordinates in `data_coords`
      
    ax : cartopy.mpl.geoaxes.GeoAxesSubplot
        If not None, `c` is plotted on the GeoAxesSubplot instance.
        Otherwise, a new figure and GeoAxesSubplot instance is created
        
    show : bool
        If True (default), the map will be showed once generated. Otherwise
        a GeoAxesSubplot instance is returned
        
    cmap : str or Colormap
        If str, it should be a valid matplotlib.cm.colormap name (see 
        matplotlib documentation). Otherwise, it should be the name of a
        colormap available in seislib.colormaps (see also the documentation at 
        https://www.fabiocrameri.ch/colourmaps/)
        
    vmin, vmax : float
        Boundaries of the colormap. If None, the minimum and maximum values
        of `c` will be taken
    
    stations_color : str
        Color of the receivers and of the great-circle paths (see matplotlib 
        documentation for valid color names). Defaults are 'r' (red) and
        'k' (black)
                
    oceans_color, lands_color : str
        Color of oceans and lands. The arguments are ignored if ax is not
        None. Otherwise, they are passed to `cartopy.feature.NaturalEarthFeature` 
        (to the argument 'facecolor'). Defaults are 'water' and 'land'

    edgecolor : str
        Color of the boundaries between, e.g., lakes and land. The argument 
        is ignored if ax is not None. Otherwise, it is passed to 
        cartopy.feature.NaturalEarthFeature (to the argument 'edgecolor'). 
        Default is 'k' (black)

    stations_alpha, paths_alpha : float, optional
        Transparency of the stations and of the great-circle paths. Defaults
        are `None` and 0.3
        
    paths_width : float
        Linewidth of the great-circle paths
                
    projection : str
        Name of the geographic projection used to create the `GeoAxesSubplot`.
        (Visit the `cartopy website 
        <https://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html>`_ 
        for a list of valid projection names.) If ax is not None, `projection` 
        is ignored. Default is 'Mercator'
        
    resolution : {'10m', '50m', '110m'}
        Resolution of the Earth features displayed in the figure. Passed to
        `cartopy.feature.NaturalEarthFeature`. Valid arguments are '110m',
        '50m', '10m'. Default is '110m'. Ignored if `ax` is not `None`.
    
    map_boundaries : list or tuple of floats, shape (4,), optional
        Lonmin, lonmax, latmin, latmax (in degrees) defining the extent of
        the map
        
    bound_map : bool
        If `True`, the map boundaries will be automatically determined.
        Ignored if `map_boundaries` is not `None`
    
    colorbar : bool
        If `True` (default), a colorbar associated with `c` is displayed on 
        the side of the map        
    
    cbar_dict : dict
        Keyword arguments passed to `matplotlib.colorbar.ColorbarBase`
    
    **kwargs
        Additional keyword arguments passed to ax.plot
    
    Returns
    -------
    If `show` is `True`
        `None` 
    Otherwise 
        `GeoAxesSubplot` instance together with an instance of 
        `matplotlib.colorbar.Colorbar` (if `colorbar` is `True`)
    """
    def get_map_boundaries(data):
        lats = np.concatenate((data[:,0], data[:, 2]))
        lons = np.concatenate((data[:,1], data[:, 3]))
        latmin, latmax = np.nanmin(lats), np.nanmax(lats)
        lonmin, lonmax = np.nanmin(lons), np.nanmax(lons)
        dlon = (lonmax - lonmin) * 0.03
        dlat = (latmax - latmin) * 0.03
        lonmin = lonmin-dlon if lonmin-dlon > -180 else lonmin
        lonmax = lonmax+dlon if lonmax+dlon < 180 else lonmax
        latmin = latmin-dlat if latmin-dlat > -90 else latmin
        latmax = latmax+dlat if latmax+dlat < 90 else latmax
        return (lonmin, lonmax, latmin, latmax)        
    
    def create_colourbar(ax, cmap, norm, **kwargs):
        """Create a colourbar with limits of lwr and upr"""
        divider = make_axes_locatable(ax)
        orientation = kwargs.pop('orientation', 'horizontal')
        if orientation == 'vertical':
            loc = 'right'
        elif orientation == 'horizontal':
            loc = 'bottom'
        cax = divider.append_axes(loc, '5%', pad='3%', axes_class=mpl.pyplot.Axes)
        cb = mpl.colorbar.ColorbarBase(cax, 
                                       cmap=cmap, 
                                       norm=norm, 
                                       orientation=orientation, 
                                       **kwargs)
        return cb

    
    if ax is None:
        projection = lower_threshold_projection(eval('ccrs.%s'%projection))
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection=projection)
        add_earth_features(ax, 
                           scale=resolution,
                           oceans_color=oceans_color, 
                           edgecolor=edgecolor,
                           lands_color=lands_color)
        
    transform = ccrs.PlateCarree()
    marker = kwargs.pop('marker', '^')
    vmin = vmin if vmin is not None else np.nanmin(c)
    vmax = vmax if vmax is not None else np.nanmax(c)
    norm = mpl.colors.Normalize(vmin, vmax)
    
    
    stations = set()
    for (lat1, lon1, lat2, lon2), v in zip(data_coords, c):
        ax.plot([lon1, lon2], 
                [lat1, lat2], 
                alpha=paths_alpha, 
                lw=paths_width, 
                color=cmap(norm(v)),
                transform=ccrs.Geodetic())
        if (lat1, lon1) not in stations:
            ax.plot(lon1, 
                    lat1, 
                    marker=marker, 
                    c=stations_color, 
                    transform=transform, 
                    alpha=stations_alpha, 
                    **kwargs) 
            stations.add((lat1, lon1))
        if (lat2, lon2) not in stations:
            ax.plot(lon2, 
                    lat2, 
                    marker=marker, 
                    c=stations_color, 
                    transform=transform, 
                    alpha=stations_alpha, 
                    **kwargs) 
            stations.add((lat2, lon2))
            
    if map_boundaries is None and bound_map:
        map_boundaries = get_map_boundaries(data_coords)
    if map_boundaries is not None:
        ax.set_extent(map_boundaries, transform)  
    if colorbar:
        cb = create_colourbar(ax, cmap, norm, **cbar_dict)    
    if show:
        plt.show()
    else:
        return ax, cb if colorbar else ax       


def plot_map(mesh, c, ax=None, projection='Mercator', map_boundaries=None, 
             bound_map=True, colorbar=True, show=True, style='colormesh', 
             add_features=False, resolution='110m', cbar_dict={}, **kwargs):
    """ 
    Utility function to display the lateral variations of some quantity on 
    a equal-area grid
    
    Parameters
    ----------
    mesh : ndarray, shape (n, 4)
        Grid cells in seislib format, consisting of n pixels described by 
        lat1, lat2, lon1, lon2 (in degrees)
        
    c : ndarray, shape (n,)
        Values associated with the grid cells (mesh)
        
    ax : cartopy.mpl.geoaxes.GeoAxesSubplot
        If not None, `c` is plotted on the `GeoAxesSubplot` instance. 
        Otherwise, a new figure and `GeoAxesSubplot` instance is created
        
    projection : str
        Name of the geographic projection used to create the `GeoAxesSubplot`.
        (Visit the `cartopy website 
        <https://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html>`_ 
        for a list of valid projection names.) If ax is not None, `projection` 
        is ignored. Default is 'Mercator'
        
    map_boundaries : list or tuple of floats, shape (4,), optional
        Lonmin, lonmax, latmin, latmax (in degrees) defining the extent of
        the map
        
    bound_map : bool
        If `True`, the map boundaries will be automatically determined.
        Ignored if `map_boundaries` is not None
    
    colorbar : bool
        If `True` (default), a colorbar associated with `c` is displayed on 
        the side of the map
        
    show : bool
        If `True` (default), the map will be showed once generated
        
    style : {'colormesh', 'contourf', 'contour'}
        Possible options are 'colormesh', 'contourf', and 'contour',
        each corresponding to the homonymous method.
        Default is 'colormesh'
        
    add_features : bool
        If `True`, natural Earth features will be added to the `GeoAxesSubplot`.
        Default is `False`. If `ax` is `None`, it is automatically set to `True`
        
    resolution : {'10m', '50m', '110m'}
        Resolution of the Earth features displayed in the figure. Passed to
        `cartopy.feature.NaturalEarthFeature`. Valid arguments are '110m',
        '50m', '10m'. Default is '110m'
        
    cbar_dict : dict
        Keyword arguments passed to `matplotlib.colorbar.ColorbarBase`
        
    **kwargs
        Additional inputs passed to the 'colormesh', 'contourf', and
        'contour' methods of :class:`SeismicTomography`
        
        
    Returns
    -------
    If show is True
        `None`
    Otherwise
        an instance of `matplotlib.collections.QuadMesh`, together with an 
        instance of `matplotlib.colorbar.Colorbar` (if `colorbar` is `True`)
    """
    def get_map_boundaries(mesh):
        latmin, lonmin = np.min(mesh, axis=0)[::2]
        latmax, lonmax = np.max(mesh, axis=0)[1::2]
        dlon = (lonmax - lonmin) * 0.03
        dlat = (latmax - latmin) * 0.03
        
        lonmin = lonmin-dlon if lonmin-dlon > -180 else lonmin
        lonmax = lonmax+dlon if lonmax+dlon < 180 else lonmax
        latmin = latmin-dlat if latmin-dlat > -90 else latmin
        latmax = latmax+dlat if latmax+dlat < 90 else latmax
        
        return (lonmin, lonmax, latmin, latmax)
    
    def set_colorbar_aspect(cb, kwargs):
        norm = kwargs.get('norm', None)
        if norm is not None and type(norm)==type(LogNorm()):
            if style in ['contour', 'contourf']:
                cmin, cmax = cb.vmin, cb.vmax
                ticks = np.power(10, np.arange(np.floor(np.log10(cmin)), 
                                               np.ceil(np.log10(cmax)+1)))
                cb.set_ticks(ticks)
                minorticks = np.hstack([np.arange(2*i, 10*i, 1*i) for i in ticks])
                minorticks = [i for i in minorticks if cmin<i<cmax]
                cb.ax.xaxis.set_ticks(minorticks, minor=True)      
        return cb

    add_features = add_features if ax is not None else True
    if ax is None:
        projection = eval('ccrs.%s()'%projection)
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection=projection)        
    if add_features:
        add_earth_features(ax, 
                           scale=resolution,
                           oceans_color='none', 
                           edgecolor='k',
                           lands_color='none')
        
    styles = ['colormesh', 'contourf', 'contour']
    if style not in styles:
        raise Exception('`style` not available. Please choose among %s'%styles)
    plotting_func = eval(style)
    
    notnan = np.flatnonzero(~np.isnan(c))
    img = plotting_func(mesh=mesh[notnan], c=c[notnan], ax=ax, **kwargs)
    
    if map_boundaries is None and bound_map:
        map_boundaries = get_map_boundaries(mesh)
    if map_boundaries is not None:
        ax.set_extent(map_boundaries, ccrs.PlateCarree())  
    if colorbar:
        cb = make_colorbar(ax, img, **cbar_dict)
        cb = set_colorbar_aspect(cb, kwargs)
    if show:
        plt.show()
    else:
        return (img, cb) if colorbar else img        


def scientific_label(obj, precision):
    """ Creates a scientific label approximating a real number in LaTex style
    
    Parameters
    ----------
    obj : float
        Number that must be represented
        
    precision : int
        Number of decimal digits in the resulting label
        
        
    Returns
    -------
    scientific_notation : str
        Approximated number, e.g., r'5.000 \\times 10^{-6}', arising from
        the function call scientific_label(5e-06, 3)
        
    
    Examples
    --------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> x = np.linspace(-2, 2, 150)
    >>> y = x**2 * 1e-5
    >>> plt.plot(x, y)
    >>> ticks = plt.gca().get_yticks()
    >>> labels = [r'$%s$'%scientific_label(i, 1) for i in ticks]
    >>> plt.yticks(ticks=ticks, labels=labels)
    >>> plt.show()
    """
    precision = '%.' + '%se'%precision
    python_notation = precision % obj
    number, power = python_notation.split('e')
    scientific_notation = r'%s \times 10^{%s}'%(number, int(power))
    return scientific_notation

















#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Feb 17 23:05:58 2019

@author: Fabrizio Magrini
@email1: fabrizio.magrini@uniroma3.it
@email2: fabrizio.magrini90@gmail.com

On terminal:
    
    python setup.py build_ext --inplace

for html:
    cython -a example_cython.pyx

"""

from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
from Cython.Build import cythonize
import numpy
import os
import platform


names = ['_math', '_spherical_geometry', '_tomography']
sources = ['%s.pyx'%name for name in names]
extra_compile_args = ["-O3", "-ffast-math", "-march=native", "-fopenmp" ]

platform_name = platform.system()


if platform_name.lower() == 'darwin':
    versions = os.listdir('/usr/local/Cellar/gcc/')
    version = max(versions, key=lambda i: int(i.split('.')[0]))
    version_int = version.split('.')[0]
    path = '/usr/local/Cellar/gcc/%s/lib/gcc/%s'%(version, version_int)
    os.environ['CC'] = 'gcc-%s'%version_int
    os.environ['CXX'] = 'g++-%s'%version_int
    extra_link_args=['-Wl,-rpath,%s'%path]

else:
    extra_link_args=['-fopenmp']


for name, source in zip(names, sources):
    language = 'c++' if name!='_math' else None
    ext_modules=[Extension(name,
                           sources=[source],
                           extra_compile_args=extra_compile_args,
                           extra_link_args=extra_link_args,
                           language=language)
                 ]
        
    setup(name=name,
          cmdclass={"build_ext": build_ext},
          ext_modules=cythonize(ext_modules, language_level="3"),
          include_dirs=[numpy.get_include()]
    )


    









#!/usr/bin/env python3
# -*- coding: utf-8 -*-
r"""
Least-Squares Imaging (Ray Theory)
==================================

To map lateral variations in surface-wave velocity, SeisLib implements a 
least-square inversion scheme based on ray theory. The travel time along the 
great-circle path can be written :math:`t = \int_{path}{s(\phi(l), \theta(l)) dl}`, 
where :math:`\phi` and :math:`\theta` denote longitude and latitude, and `s` the sought 
Earth's slowness.

Let us consider a discrete parameterization of the Earth's surface, and assume each block 
(or grid cell) of such parameterization has constant slowness. The above integral expression 
can then be reformulated in the discrete form

.. math::

    s = \frac{1}{L} \sum_{n}{s_n l_n},

where `L` is the length of the great-circle path and `l` the distance traveled by the surface 
wave through the :math:`n`\ th block. The above equation represents the forward calculation that allows 
for retrieving the average velocity of propagation between two points on the Earth's surface, 
provided that the (discrete) spatial variations in velocity (or slowness) are known. If we now 
define the :math:`m \times n` matrix such that :math:`A_{ij} = \frac{l_j}{L_i}`, where :math:`L_i` 
is the length of the great circle associated with :math:`i`\ th observation, we can switch to matrix 
notation and write

.. math::

    \bf A \cdot x = d,

where :math:`\bf d` is an `m`-vector whose :math:`k`\ th element corresponds to the measured slowness, 
and :math:`\bf x` the sought `n`-vector whose :math:`k`\ th element corresponds to the model 
coefficient :math:`s_k`. Matrix :math:`\bf A` can be computed numerically in a relatively simple 
fashion. In real-world seismological applications, however, the above system of equations 
is often strongly overdetermined, i.e. the number of data points is much larger than the number 
of model parameters (:math:`m \gg n`). This implies that, although :math:`\bf A` is known, it is 
not invertible.

SeisLib solves the above inverse problem in a regularized least-squares sense [1]_, i.e.

.. math::

    {\bf x} = {\bf x}_0 + \left( {\bf A}^T \cdot {\bf A} + \mu^2 {\bf R}^T \cdot {\bf R} \right)^{-1} 
    \cdot 
    {\bf A}^T \cdot ({\bf d - A} \cdot {\bf x}_0),

where the roughness operator :math:`\bf R` is dependent on the parameterization, 
:math:`{\bf x}_0` is a reference model, and the scalar weight :math:`\mu` should be chosen via 
L-curve analysis [2]_.


References
----------
.. [1] Aster et al. 2018, Parameter estimation and inverse problems
.. [2] Hansen 2001, The L-curve and its use in the numerical treatment of inverse problems

"""
import pickle
import warnings
import numpy as np
import scipy.sparse
import matplotlib.pyplot as plt
#from scipy.linalg import pinvh, lstsq
from obspy.geodetics import gps2dist_azimuth
from seislib.tomography import EqualAreaGrid, RegularGrid
from seislib import colormaps as scm
import seislib.plotting as plotting
from seislib.tomography._ray_theory._tomography import _compile_coefficients
from seislib.tomography._ray_theory._tomography import _refine_parameterization
from seislib.tomography._ray_theory._tomography import _raypaths_per_pixel
from seislib.tomography._ray_theory._tomography import _derivatives_lat_lon
from seislib.tomography._ray_theory._tomography import _select_parameters
eps = np.finfo(np.float64).eps


class SeismicTomography:
    r""" 
    Class to obtain velocity maps using a linearized inversion based on the
    ray theory (infinite frequency approximation and wave propagation along
    the great-circle path connecting two points on the Earth's surface).
    
    Parameters
    ----------
    cell_size : int
        Size of each side of the equal-area grid
        
    latmin, lonmin, latmax, lonmax : float, optional
        Boundaries (in degrees) of the grid
        
    regular_grid : bool
        If False (default), the study area is discretized using an equal-area
        parameterization. Otherwise, a regular grid is employed.
        
    verbose : bool
        If `True`, information about the grid and data will be displayed in
        console. (Default is `True`)


    Attributes
    ----------
    verbose : bool
        If `True` (default), information about the grid and data will be displayed 
        in console
        
    grid : seislib.EqualAreaGrid
        Instance of :class:`seislib.tomography.grid.EqualAreaGrid`, corresponding 
        to an equal-area parameterization
    
    lonmin_data, latmin_data, lonmax_data, latmax_data : float
        Minimum and maximum values of latitude and longitude in the data.
        Only available after function call :meth:`add_data`
    
    data_coords : ndarray of shape (n, 4)
        Lat1, lon1, lat2, lon2 of the pairs of receivers (or epicenter-receiver),
        in degrees (-180<lon<180, -90<lat<90). Only available after function 
        call :meth:`add_data`
        
    velocity : ndarray of shape (n,) 
        Velocity (in m/s) measured between each pair of stations or 
        epicenter-station. Only available after function call :meth:`add_data`
        
    refvel : float
        Reference velocity (in m/s), used in the inversion. Only available after 
        function call :meth:`add_data`
        
    A : ndarray of shape (n, m)
        Where n is the number of data, m is the number of parameters (i.e.,
        number of rows in self.grid.mesh). Only available after function call 
        :meth:`compile_coefficients`
            
    
    Examples
    --------
    The following will calculate a phase-velocity map at a given period, say
    10 s, based on inter-station measurements of surface-wave velocity.
    In practice, our data consist of a ndarray of shape (n, 5), where n is the
    number of inter-station measurements of phase velocity (extracted, for
    example, via :class:`seislib.an.an_velocity.AmbientNoiseVelocity`), and the 
    five columns consist of lat1 (°), lon1 (°), lat2 (°), lon2 (°), velocity (m/s), 
    respectively. (-180<=lon<180, -90<=lat<90). This matrix has been saved to 
    `/path/to/data.txt`
    
    We will discretize the study area using an equal-area parameterization, 
    characterized by blocks of :math:`2^{\circ} \times 2^{\circ}. We will then
    iteratively refine the parameterization up to a maximum number of 2 times, 
    to reach a maximum resolution of 0.5° in the areas of the map characterized 
    by a relatively large number of measurements. (This refinement can be 
    carried out an arbitrary number of times.)
    
    First, we need to initialize the :class:`SeismicTomography` instance and 
    load our data into memory:
    
    >>> from seislib.tomography import SeismicTomography
    >>> tomo = SeismicTomography(cell_size=2, regular_grid=False, verbose=True)
    >>> tomo.add_data(src='/path/to/data')
    -------------------------------------
    Optimal grid found in 46 iterations
    -------------------------------------
    -------------------------------------
    GRID PARAMETERS
    Lonmin - Lonmax : -180.000 - 180.000
    Latmin - Latmax : -90.000 - 90.000
    Number of cells : 10312
    Grid cells of 2.000° : 10312
    -------------------------------------
    DATA PARAMETERS
    Lonmin - Lonmax data : -124.178 - -69.292
    Latmin - Latmax data : 23.686 - 48.470
    Number of measurements : 1064
    Source : /path/to/data
    -------------------------------------        
    
    .. hint:: 
        we could have directly passed the data matrix without
        loading it from disk. If `data` is your ndarray of shape (m, 5),
        you can pass it to `tomo` by::
        
            tomo.add_data(data=your_matrix)

    .. hint:: 
        you can add, sequentially, how many data sets you wish. The data
        information will be automatically updated

    .. hint:: 
        to display general information on the data, type `print(tomo)`. 
        To display general information on the parameterization, type 
        `print(tomo.grid)`. (See also the documentation on 
        :class:`seislib.tomography.grid.EqualAreaGrid`)
        
    .. hint:: 
        if you are interested in working at local scale, where the use of 
        equal-area parameterizations does not have clear advantages, consider
        setting `regular_grid=True` when initializing the instance of 
        :class:`SeismicTomography`, together with the boundaries of the region 
        you are interested in (`lonmin`, `lonmax`, `latmin`, `latmax`). For 
        example::
            
            from seislib.tomography import SeismicTomography
            
            tomo = SeismicTomography(cell_size=0.05, 
                                     regular_grid=True, 
                                     latmin=40,
                                     latmax=41,
                                     lonmin=10,
                                     lonmax=12,
                                     verbose=True)
            
            
    Now we can restrict the boundaries of the (global) equal-area 
    parameterization to the minimum and maximum latitude and longitude 
    spanned by our data:
        
    >>> tomo.grid.set_boundaries(latmin=tomo.latmin_data,
                                 latmax=tomo.latmax_data,
                                 lonmin=tomo.lonmin_data,
                                 lonmax=tomo.lonmax_data)
    *** GRID UPDATED ***
    -------------------------------------
    GRID PARAMETERS
    Lonmin - Lonmax : -126.234 - -66.614
    Latmin - Latmax : 22.006 - 50.005
    Number of cells : 324
    Grid cells of 2.000° : 324
    -------------------------------------
    
    Having done so, everything is ready to calculate the coefficients of the
    :math:`\bf A` matrix (i.e., of the data kernel), and to refine the 
    parameterization up to two times in the areas characterized by a relatively 
    high density of measurements. We will define such regions (i.e., model 
    parameters) as those where there are at least 150 inter-station great-circle 
    paths intersecting them. In doing so, we will remove the grid cells of 2° 
    that are not intersected by at least one great-circle path (see the argument 
    `keep_empty_cells`).
        
    >>> tomo.compile_coefficients(keep_empty_cells=False)
    >>> tomo.refine_parameterization(hitcounts=150, 
    ...                              keep_empty_cells=True)    
    >>> tomo.refine_parameterization(hitcounts=150, 
    ...                              keep_empty_cells=True)    
    *** GRID UPDATED ***
    -------------------------------------
    GRID PARAMETERS
    Lonmin - Lonmax : -126.142 - -66.614
    Latmin - Latmax : 22.006 - 50.005
    Number of cells : 202
    Grid cells of 2.000° : 202
    -------------------------------------
    *** GRID UPDATED ***
    -------------------------------------
    GRID PARAMETERS
    Lonmin - Lonmax : -126.142 - -66.614
    Latmin - Latmax : 22.006 - 50.005
    Number of cells : 808
    Grid cells of 2.000° : 0
    Grid cells of 1.000° : 808
    -------------------------------------
    *** GRID UPDATED ***
    -------------------------------------
    GRID PARAMETERS
    Lonmin - Lonmax : -126.142 - -66.614
    Latmin - Latmax : 22.006 - 50.005
    Number of cells : 2284
    Grid cells of 2.000° : 0
    Grid cells of 1.000° : 316
    Grid cells of 0.500° : 1968
    -------------------------------------
                
    To obtain the velocity map, we now need to carry out the inversion. We will
    apply a roughness-damping regularization, using a roughness coefficient
    equal to 3e-3 (to select a proper roughness damping, check
    :meth:`lcurve`). We then plot the retrieved velocity::
        
        c = 1 / tomo.solve(rdamp=3e-3)
        tomo.plot_map(tomo.grid.mesh, c)
        
    (Note that the `solve` method returns slowness, hence we took the inverse
    of the solution.) A checkerboard test, to verify the ability of our data to 
    resolve the lateral differences in the study area, can simply be performed
    and visualized by::
    
        restest = tomo.checkerboard_test(kx=6,
                                         ky=6,
                                         latmin=tomo.grid.latmin,
                                         latmax=tomo.grid.latmax,
                                         lonmin=tomo.grid.lonmin,
                                         lonmax=tomo.grid.lonmax,
                                         cell_size=0.5,
                                         anom_amp=0.1,
                                         noise=0,
                                         rdamp=3e-3)
        
        input_model = 1 / restest['synth_model']
        input_mesh = restest['mesh']
        retrieved_model = 1 / restest['retrieved_model']
        
        tomo.plot_map(tomo.grid.mesh, retrieved_model)
        tomo.plot_map(input_mesh, input_model)

    
    .. note:: 
        The resolution test (check also :meth:`resolution_test` and
        :meth:`spike_test`) should be performed after (and only after) the 
        coefficients in the data-kernel (`tomo.A` matrix) have been complied, 
        i.e., after having called :meth:`compile_coefficients` (and eventually
        :meth:`refine_parameterization`, which updates `tomo.A`).
    
    
    Finally, we show how we can obtain a heat map of the raypaths intersecting
    each model parameter::
        
        raypaths = tomo.raypaths_per_pixel()
        img, cb = tomo.plot_map(mesh=tomo.grid.mesh, 
                                c=raypaths, 
                                cmap='cividis', 
                                style='contourf', 
                                levels=20,
                                show=False)
        cb.set_label('Raycounts')
        
    .. hint:: 
        When the above argument `show` is `False`, the image and colorbar will be
        returned. That allow us to change the label on the colorbar.

    .. hint:: 
        For more control on the plots, consider creating your own instance 
        of `GeoAxesSubplot` (see cartopy documentation) before calling the
        plot_map function (or, alternatively, :meth:`colormesh`, :meth:`contour`,
        or :meth:`contourf` of SeismicTomography). For example::
            
            from seislib.plotting import make_colorbar
            
            fig = plt.figure(figsize=(12, 12))
            ax = fig.add_subplot(1, 1, 1, projection=ccrs.Mercator())
            ax.coastlines(resolution='50m', color='k', lw=1, zorder=100)
            img = tomo.colormesh(mesh=tomo.grid.mesh, 
                                c=raypaths, 
                                ax=ax, 
                                cmap='cividis', 
                                shading='flat', 
                                edgecolors='face')
            map_boundaries = (tomo.grid.lonmin, tomo.grid.lonmax, 
                            tomo.grid.latmin, tomo.grid.latmax)
            ax.set_extent(map_boundaries, ccrs.PlateCarree())  
            cb = make_colorbar(ax, img, orientation='horizontal')
            cb.set_label(label='Raycounts', labelpad=10, fontsize=22)
    """ 
    
    def __init__(self, cell_size=4, latmin=None, lonmin=None, latmax=None, 
                 lonmax=None, verbose=True, regular_grid=False):    
        self.verbose = verbose
        grid = EqualAreaGrid if not regular_grid else RegularGrid
        self.grid = grid(cell_size=cell_size,
                         latmin=latmin,
                         lonmin=lonmin, 
                         latmax=latmax,
                         lonmax=lonmax,
                         verbose=verbose)

               
    def __repr__(self):
        return str(self)
              
    
    def __str__(self):
        string = 'DATA PARAMETERS\n'
        if not 'data_coords' in self.__dict__:
            return string + 'No data available'
        string += 'Lonmin - Lonmax data : %.3f - %.3f\n'%(self.lonmin_data,
                                                          self.lonmax_data)
        string += 'Latmin - Latmax data : %.3f - %.3f\n'%(self.latmin_data,
                                                          self.latmax_data)
        string += 'Number of measurements : %s\n'%self.velocity.size
        string += 'Source : %s\n'%self.src
        string += '-------------------------------------\n'
        return string
    
    
    def save(self, path):
        """ 
        Saves the SeismicTomography instance to the specified path in the pickle
        compressed format.
        
        Parameters
        ----------
        path : str
            Absolute path of the resulting file
        """
        
        if not path.endswith('.pickle'):
            path += '.pickle'    
        with open(path, 'wb') as f:
            pickle.dump(self, f)
            
            
    @classmethod
    def load(cls, path):
        """
        Loads the pickle file at the specified path and returns the associated
        instance of SeismicTomography.
        
        Parameters
        ----------
        path : str
            Absolute path of the file
        """
        
        with open(path, 'rb') as f:
            return pickle.load(f)
    
    
    def update_data_info(self, refvel=None):
        """ 
        Updates the information on minimum / maximum longitude and latitude of
        the data and the reference velocity
        
        Parameters
        ----------
        refvel : float, optional
            If not None, the reference velocity is updated along with minimum
            and maximum longitudes and latitudes
        """
        
        self.refvel = np.mean(self.velocity) if refvel is None else refvel
        self.latmin_data = min(np.min(self.data_coords[:, (0, 2)], axis=0))
        self.lonmin_data = min(np.min(self.data_coords[:, (1, 3)], axis=0))
        self.latmax_data = max(np.max(self.data_coords[:, (0, 2)], axis=0))
        self.lonmax_data = max(np.max(self.data_coords[:, (1, 3)], axis=0))
        if self.verbose:
            print(self)
    
    
    def add_data(self, data=None, refvel=None, src=None, **kwargs):
        """Loads the data to be used for tomography.

        If the argument `data` isn't passed in, the data are loaded from `src`.
        If the instance of :class:`SeismicTomography` already contains data, 
        the new data will be added (concatenated) to the previously loaded ones.
        
        .. note:: 
            In the current implementation, the 6th column of the data (i.e., 
            the standard deviation), even if present, is not employed in the 
            inversion (see :meth:`solve`). This behaviour will be changed
            in a future release.
        

        Parameters
        ----------
        data : ndarray, optional
            The array must be of shape (n, 5) or (n, 6). Cols 1 to 4 must contain
            the coordinates of the measurements (lat1, lon1, lat2, lon2), col 5
            the velocity, and col 6 the standard deviation (optional). 
            (-180<lon<180, -90<lat<90, velocity in m/s)
            
        refvel : float, optional
            Reference velocity (in m/s). If `None`, the average velocity is used as 
            reference
            
        src : str, optional
            If `data` is None, `src` is used to load the data. The file extension
            must be .txt
            
        **kwargs : optional
            Arguments to be passed to numpy.loadtxt
        
        
        Returns
        -------
        None
            The data are stored in `self.__dict__` and can be accessed through 
            `self.data_coords`, `self.velocity`, `self.refvel`, `self.std` (if 
            present), `self.latmin_data`, `self.lonmin_data`, `self.latmax_data`, 
            `self.lonmax_data`
        """
        if data is None:
            data = np.loadtxt(src, **kwargs)
        else:
            data = data.astype(np.float64)
        if 'velocity' not in self.__dict__:
            self.src = 'Unknown' if src is None else src
            self.data_coords = data[:, :4]
            self.velocity = data[:, 4]
            if data.shape[1] == 6:
                self.std = data[:, 5] 
        else:
            self.src += ' + Unknown' if src is None else ' + %s'%src
            if data.shape[1] == 6:
                if 'std' in self.__dict__:
                    self.std = np.concatenate((self.std, data[:, 5]))
                else:
                    self.std = np.concatenate((np.zeros(self.velocity.shape),
                                               data[:, 5]))
            else:
                if 'std' in self.__dict__:
                    self.std = np.concatenate((self.std, 
                                               np.zeros(self.velocity.shape)))
            self.data_coords = np.row_stack((self.data_coords, data[:, :4]))
            self.velocity = np.concatenate((self.velocity, data[:, 4]))
        
        self.update_data_info(refvel=refvel)
        

    @classmethod
    def delay_to_velocity(cls, lat1, lon1, lat2, lon2, delay, refvel):
        """ Converts a delay to velocity
        
        A negative delay `dt` corresponds to faster velocity with respect 
        to the reference velocity `v0`::

            dt = x/v - x/v0
            v = x / (dt + x/v0)

        where `x` denotes distance and `v` the observed velocity.
        
        Parameters
        ----------
        lat1, lon1, lat2, lon2 : float or array-like of shape (n,)
            Coordinates (in radians) associated with the dalay
            
        delay : float or array-like of shape (n,)
            Measured delay (in seconds)
            
        refvel : float
            Reference velocity used to calculate the delay (in m/s)
            
            
        Returns
        -------
        float or array-like of shape (n,)
        """
        
        dist = np.asarray(cls.gc_distance(lat1, lon1, lat2, lon2))
        return dist / (delay + dist/refvel)
        
    
    @classmethod
    def velocity_to_delay(cls, lat1, lon1, lat2, lon2, velocity, refvel):
        """ Converts a velocity to delay
        
        Velocities (`v`) faster than the reference (`v0`) correspond to 
        negative delays (`dt`)::

            dt = x/v - x/v0

        where `x` denotes distance.
        
        Parameters
        ----------
        lat1, lon1, lat2, lon2 : float or array-like of shape (n,)
            Coordinates (in radians) associated with the velocity
            
        velocity : float or array-like of shape (n,)
            Measured velocity (in m/s)
            
        refvel : float
            Reference velocity to be used to calculate the delay (in m/s)
            
            
        Returns
        -------
        float or array-like of shape (n,)
        """   
        
        dist = np.asarray(cls.gc_distance(lat1, lon1, lat2, lon2))
        return dist/velocity - dist/refvel
    
    
    @classmethod
    def gc_distance(cls, lat1, lon1, lat2, lon2):
        """ 
        Calculates the great circle distance (in m) between coordinate points 
        (in degrees). This function calls directly the obspy `gps2dist_azimuth`,
        it only extends its functionality through the `numpy.vectorize` decorator.
        
        Parameters
        ----------
        lat1, lon1, lat2, lon2 : float or array-like of shape (n,)
            Coordinates of the points on the Earth's surface, in degrees.        
        
        Returns
        -------
        Great-circle distance (in m) 
            If the input is an array (or list) of coordinates, an array of 
            distances is returned
        """
        
        func = np.vectorize(gps2dist_azimuth)
        return func(lat1, lon1, lat2, lon2)[0]
    
    
    @classmethod
    def azimuth_backazimuth(cls, lat1, lon1, lat2, lon2):
        """ 
        Calculates the azimuth and backazimuth (in degrees) between coordinate 
        points (in degrees). This function calls directly the obspy 
        `gps2dist_azimuth`, it only extends its functionality through the 
        numpy.vectorize decorator
        
        Parameters
        ----------
        lat1, lon1, lat2, lon2 : float or array-like of shape (n,)
            Coordinates of the points on the Earth's surface, in degrees.               
        
        Returns
        -------
        tuple of shape (2,) or ndarray of shape (n, 2)
            The columns consist of azimuth azimuth and backazimuth
        """
        
        func = np.vectorize(gps2dist_azimuth)
        return func(lat1, lon1, lat2, lon2)[1:]
    
    
    def compile_coefficients(self, refine=False, coeff_matrix=None, 
                             keep_empty_cells=True):
        """ 
        Compiles the matrix of coefficients (`A`) used to perform the inversion. 
        It calls the `_compile_coefficients` function, written in cython.
        
        
        Parameters
        ----------
        refine : bool
            should be `True` if the user has refined the parameterization of the
            mesh (by calling `refine_parameterization`). This allows to reduce
            computation time. (Default is `False`)
            
        coeff_matrix : if `None`, a new `A` will be created and stored in the
            SeismicTomography instance (`self.A`). `A` should be passed
            in after the parameterization refinement. (Default is `None`)
            
        keep_empty_cells : bool
            If `False`, cells intersected by no raypaths are removed
        
        
        Returns
        -------
        None 
            The :class:`SeismicTomography` instance will be updated, so that 
            `A` can be accessed typing `self.A`
        """
        mesh_latmax = np.radians(self.grid.latmax)
        mesh_lonmax = np.radians(self.grid.lonmax)
        mesh = np.radians(self.grid.mesh)
        data_coords = np.radians(self.data_coords)
        self.A = _compile_coefficients(data_coords=data_coords,
                                       mesh=mesh, 
                                       mesh_latmax=mesh_latmax, 
                                       mesh_lonmax=mesh_lonmax,
                                       refine=refine,
                                       coeff_matrix=coeff_matrix)
        if not keep_empty_cells:
            raycounts = self.raypaths_per_pixel()
            keep_pixels = np.flatnonzero(raycounts > 0).astype(np.int32)
            self.A = _select_parameters(self.A, keep_pixels)
            self.grid.select_cells(indexes=keep_pixels)
        
    
    def refine_parameterization(self, hitcounts=100, keep_empty_cells=True,
                                latmin=None, latmax=None, lonmin=None, lonmax=None):
        """ 
        Halves the dimension of the pixels intersected by a number of 
        raypaths >= hitcounts.        
        
        Parameters
        ----------
        hitcounts : int        
            Each parameter (grid cell) intersected by a number of raypaths
            equal or greater than this threshold is "refined", i.e., splitted
            in four equal-area sub-parameters. Default is 100
            
        keep_empty_cells : bool
            If `False`, cells intersected by no raypaths are removed
            
        lat1, lon1, lat2, lon2 : float
            If specified (in degrees), only the region inside this boundary is 
            considered for refining the parameterization. (-180<lon<180, 
            -90<lat<90)


        Returns
        -------
        None
            `self.grid.mesh` and `self.A` are updated
        """
        mesh = np.radians(self.grid.mesh)
        if any([latmin, latmax, lonmin, lonmax]):
            latmin = latmin if latmin is not None else -90
            latmax = latmax if latmax is not None else +90
            lonmin = lonmin if lonmin is not None else -180
            lonmax = lonmax if lonmax is not None else +180
            region_to_refine = np.radians([latmin, latmax, lonmin, lonmax])
        else:
            region_to_refine = None

        newmesh, A = _refine_parameterization(mesh, 
                                              self.A, 
                                              hitcounts=hitcounts,
                                              region_to_refine=region_to_refine)
        self.grid.update_grid_params(np.degrees(newmesh), refined=True)
        self.compile_coefficients(refine=True, coeff_matrix=A, 
                                  keep_empty_cells=keep_empty_cells)

    
    def raypaths_per_pixel(self):
        """ 
        Calculates the number of raypaths in each pixel of the mesh stored in
        the :class:`SeismicTomography` instance (`self.grid.mesh`)
        
        Returns
        -------
        ndarray of dimension self.grid.mesh.shape[0]            
        """
        if 'A' not in self.__dict__:
            self.compile_coefficients()
        return np.asarray(_raypaths_per_pixel(self.A))
                    
    
    def reduce_measurements(self, latmin=-90, latmax=90, lonmin=-180, lonmax=180):
        """
        Any measurement not intersecting the specified region is removed from
        the data. Consider using this function in presence of large datasets,
        for filtering out those rays that do not constrain the parameters (i.e.,
        the grid cells) of interest in the inversion.
        
        Parameters
        ----------
        lat1, lon1, lat2, lon2 : float (in degrees)
            Boundaries of the region of interest. (-180<lon<180, -90<lat<90)
        
        
        Returns
        -------
        None
            The instance of :class:`SeismicTomography` is updated with the new 
            data

        """
        latmin = latmin if latmin is not None else -90
        latmax = latmax if latmax is not None else +90
        lonmin = lonmin if lonmin is not None else -180
        lonmax = lonmax if lonmax is not None else +180
        mesh = self.grid.mesh
        ilat = np.flatnonzero((mesh[:,0]<=latmax) & (mesh[:,1]>=latmin))
        ilon = np.flatnonzero((mesh[:,2]<=lonmax) & (mesh[:,3]>=lonmin))
        ipixels = np.intersect1d(ilat, ilon)
        if 'A' not in self.__dict__:
            self.compile_coefficients(keep_empty_cells=True)
        idata = [i for i in range(self.A.shape[0]) if np.any(self.A[i, ipixels])]
        self.data_coords = self.data_coords[idata]
        self.velocity = self.velocity[idata]
        del self.A
        self.update_data_info()
    
    
    def solve(self, A=None, slowness=None, refvel=None, mesh=None, ndamp=0, 
              rdamp=0):#, return_resmatrix=False):
        r"""
        Method for solving the (regularized) inverse problem Ax = b, 
        where A is the matrix of coefficients, x the 
        sought model of slowness, and b the array of measurements 
        expressed in slowness. The method first computes
        the array of residuals r = b - Ax0, where x0 denotes the reference 
        slowness passed to the function as `refvel` (and converted to
        slowness prior to the inversion). The residuals are then used to retrieve
        the least square solution 

        .. math::
            {\bf x}_{LS} = ({\bf A^{T}A} + \mu^2 {\bf I} + \rho^2 {\bf I G})^-1 
            \bf{A^{T} r},

        where `T` denotes transposition, I the identity matrix, and
        :math:`\mu` and :math:`\rho` are norm and roughness damping, respectively. 
        The square matrix G has the same dimensions of the model x. The above solution 
        is then used to compute the final model :math:`{\bf x = x0 + x}_{LS}`.
        
        Parameters
        ----------
        A : ndarray, shape (m, n), optional
            Matrix of the coefficients. If None (default), the `A` matrix
            stored in the :class:`SeismicTomography` instance is used
            
        slowness : ndarray, shape (m,), optional
            Inverse of the velocity measurements. Corresponds to b in the 
            equation Ax = b. If None (default), the velocity stored in the 
            :class:`SeismicTomography` instance is used to compute slowness
            
        refvel : float, optional
            Reference velocity (in m/s) used for the initial guess of the velocity
            model. If provided, should stabilize the inversion and make it 
            converge faster. If None (default), the reference velocity stored
            in the :class:`SeismicTomography` instance is used
            
        mesh : ndarray, shape (n, 4), optional
            Grid cells in seislib format, consisting of n pixels described by 
            lat1, lat2, lon1, lon2 (in degrees). If None (default), the mesh 
            stored in the :class:`SeismicTomography` instance is used
            
        ndamp : float
            Norm damping coefficient (default is zero)
            
        rdamp : float
            Roughness damping coefficient (default is zero)
            
        
        Returns
        -------
        ndarray of shape (n,)
            Least-square solution (slowness, in s/m)
        """
        
        def norm_damping(mesh, damp):
            I = np.identity(mesh.shape[0])
            I *= np.diff(mesh[:, :2])**2
            return scipy.sparse.csr_matrix(I * damp**2)
        
        def roughness_damping(mesh, damp):
            G_lat, G_lon = _derivatives_lat_lon(mesh)
            G_lat = scipy.sparse.csr_matrix(G_lat)
            G_lon = scipy.sparse.csr_matrix(G_lon)
            return damp**2 * (G_lat.T.dot(G_lat) + G_lon.T.dot(G_lon))
                        
        A = A if A is not None else self.A
        if not isinstance(A, scipy.sparse.csr_matrix):
            A = scipy.sparse.csr_matrix(A)
        slowness = slowness if slowness is not None else 1/self.velocity
        refvel = refvel if refvel is not None else self.refvel
        mesh = np.radians(mesh) if mesh is not None else np.radians(self.grid.mesh)
        x0 = np.ones(A.shape[1]) * 1/refvel
        residuals = slowness - A.dot(x0)
        lhs = A.T @ A
        if ndamp > 0:
            lhs += norm_damping(mesh=mesh, damp=ndamp)
        if rdamp > 0:
            lhs += roughness_damping(mesh=mesh, damp=rdamp)
        rhs = A.T.dot(residuals)  
#        x = x0 + scipy.linalg.solve(lhs.todense(), rhs)
#        if return_resmatrix:
#            resmatrix = scipy.linalg.inv(lhs.todense()) @ A.T @ A
#            return x, resmatrix
        return x0 + scipy.linalg.solve(lhs.todense(), rhs)

    
    def lcurve(self, A=None, slowness=None, refvel=None, mesh=None, damping='roughness', 
               n=20, damp_min=1e-5, damp_max=1e-1, logspace=True, show=True):
        """ L-curve analysis. The function calls iteratively the `solve` method
        
        Parameters
        ----------
        A : ndarray, shape (m, n), optional
            Matrix of the coefficients. If `None` (default), the `A` matrix
            stored in the :class:`SeismicTomography` instance is used
            
        slowness : ndarray, shape (m,), optional
            Inverse of the velocity measurements. Corresponds to b in the 
            equation Ax = b. If `None` (default), the velocity stored in the 
            :class:`SeismicTomography` instance is used to compute slowness
            
        refvel : float, optional
            Reference velocity (in m/s) used for the initial guess of the velocity
            model. If provided, should stabilize the inversion and make it 
            converge faster. If `None` (default), the reference velocity stored
            in the :class:`SeismicTomography` instance is used
            
        mesh : ndarray, shape (n, 4), optional
            Grid cells in seislib format, consisting of n pixels described by 
            lat1, lat2, lon1, lon2. If `None` (default), the mesh stored in the 
            :class:`SeismicTomography` instance is used
            
        damping : {'norm', 'roughness'}
            Damping criterion used in the inversion. If 'norm', a norm damping
            is applied, otherwise the inversion is regularized using roughness
            damping (default)
            
        n : int
            Number of inversions performed in the analysis
            
        damp_min : float
            Minimum damping term (default is 1e-5)
            
        damp_max : float
            Maximum damping term (default is 1e-1)
            
        logspace : bool
            If `True` (default), the damping range is defined to be linearly 
            spaced in logarithmic scale between `damp_min` and `damp_max`, i.e.,
            `numpy.logspace(np.log10(damp_min), np.log10(damp_ax), n)`. `False` 
            corresponds to `numpy.linspace(damp_min, damp_max, n)`
            
        show : bool
            If `True` (default), the result of the analysis is displayed
        
        Returns
        -------
        damp_range : ndarray (n,)
            Damping range used in the analysis
            
        result : tuple of shape (2,)
            (i) Residual norm |Ax - b| and (ii) norm of the damped solution 
            |Gx|, where G = I (identity matrix) if the L-curve was performed 
            using a norm damping (`ndamp`>0), or the roughness operator if the 
            roughness damping was used instead (`rdamp`>0)
        """
        
        def plot(dampings, models, ylabel, scale):
            plt.figure(figsize=(10, 8))
            for damping, (residual_norm, yaxis) in zip(dampings, models):
                plt.plot(residual_norm, yaxis, 'ro')
                plt.text(residual_norm - residual_norm*2e-3, 
                         yaxis - yaxis*2e-3, 
                         s=r'$%s$'%plotting.scientific_label(damping, 1), 
                         va='top', 
                         ha='right', 
                         fontsize=11,
                         color='r')
            plt.xlabel(r'$||$ A$\cdot$x - b $||$')
            plt.ylabel(ylabel, fontsize=20, labelpad=10)
            plt.yscale(scale)
            plt.xscale(scale)
            plt.show()
        
        A = A if A is not None else self.A
        if not isinstance(A, scipy.sparse.csr_matrix):
            A = scipy.sparse.csr_matrix(A)
        slowness = slowness if slowness is not None else 1/self.velocity
        refvel = refvel if refvel is not None else self.refvel
        mesh = mesh if mesh is not None else self.grid.mesh
        if damping!='roughness' and damping!='norm':
            string = 'Unrecognized damping type: only "roughness" and "norm"'
            string += ' are allowed. Roughness damping will be used instead.'
            warnings.warn(string)
        if logspace:
            damp_range = np.logspace(np.log10(damp_min), np.log10(damp_max), n)
        else:
            damp_range = np.linspace(damp_min, damp_max, n)
        result = []
        if damping == 'roughness':
            G_lat, G_lon = _derivatives_lat_lon(mesh)
            G_lat = scipy.sparse.csr_matrix(G_lat)
            G_lon = scipy.sparse.csr_matrix(G_lon)
            G = (G_lat.T.dot(G_lat) + G_lon.T.dot(G_lon))
        for damp in damp_range:
            if damping == 'roughness':
                ndamp = 0
                rdamp = damp
            else:
                ndamp = damp
                rdamp = 0     
            x = self.solve(A=A, slowness=slowness, mesh=mesh, ndamp=ndamp, 
                           rdamp=rdamp, refvel=refvel)
            residual_norm = np.linalg.norm(slowness - A.dot(x))
            yaxis = np.linalg.norm(x) if ndamp else np.linalg.norm(G.dot(x))
            result.append((residual_norm, yaxis))
        if show:
            ylabel = r'$||$ G $\cdot$ x $||$' if damping=='roughness' else r'$||$ x $||$'
            scale = 'log' if logspace else 'linear'
            plot(damp_range, result, ylabel, scale)
            
        return damp_range, result
    
        
    @staticmethod
    def checkerboard(ref_value, kx=10, ky=10, latmin=None, latmax=None, lonmin=None, 
                     lonmax=None, anom_amp=0.1):
        """ Checkerboard-like pattern on Earth
        
        Parameters
        ----------
        ref_value : float
            Reference value with respect to which the anomalies are generated
            
        kx, ky : float
            Frequency of anomalies along longitude (kx) and latitude (ky) within
            the boundaries defined by `lonmin`, `lonmax`, `latmin`, `latmax`.
            Defaults are 10.
            
        latmin, latmax, lonmin, lonmax : float, optional
            Boundaries of the checkerboard (in degrees). 
            
        anom_amp : float
            Amplitude of the anomalies. Default is 0.1 (i.e. 10% of the reference
            value)
        
        
        Returns
        -------
        Function
            Generating function that should be called passing the values of longitude
            and latitude where the values of the checkerboard want to be retrieved.
        
        
        Examples
        --------
        >>> from seislib.tomography import SeismicTomography
        >>> tomo = SeismicTomography(5)
        >>> lat = np.mean(tomo.grid.mesh[:, :2], axis=1)
        >>> lon = np.mean(tomo.grid.mesh[:, 2:], axis=1)
        >>> pattern = tomo.checkerboard(ref_value=3, kx=6, ky=6)
        >>> c = pattern(lon, lat)
        >>> img, cb = tomo.plot_map(tomo.grid.mesh, c, projection='Robinson', 
        ...                         show=False)
        >>> cb.set_label('Checkerboard Pattern')
        """
        def generating_function(lon, lat):
            smooth_poles = np.cos(np.radians(lat))
            lon_pattern = np.sin(x_spatial_freq * lon)
            lat_pattern = np.sin(y_spatial_freq * lat)
            values = (lon_pattern + lat_pattern) * smooth_poles
            values *= anom_amp / np.max(np.abs(values))
            if ref_value is not None:
                return ref_value + values*ref_value
            return values
            

        latmin = latmin if latmin is not None else -90
        lonmin = lonmin if lonmin is not None else -180
        latmax = latmax if latmax is not None else 90
        lonmax = lonmax if lonmax is not None else 180       
        x_spatial_freq = kx * 2 * np.pi / (lonmax - lonmin)
        y_spatial_freq = ky * 2 * np.pi / (latmax - latmin)
        
        return generating_function
            
  
    @staticmethod
    def spike(ref_value, x0, y0, sigma_x=1, sigma_y=1, anom_amp=0.1):
        """ Spike-like pattern on Earth, created via a 3-D Gaussian
        
        Parameters
        ----------
        ref_value : float
            Reference value with respect to which the anomalies are generated
            
        x0, y0 : float
            Central longitude and latitude of the anomaly
            
        sigma_x, sigma_y : float
            Standard deviations in the longitude and latitude directions of the
            3-D Gaussian. Default is 1
            
        anom_amp : float
            Amplitude of the anomalies. Default is 0.1 (i.e. 10% of the reference
            value)
            
        Returns
        -------
        Function
            Generating function that should be called passing the values of longitude
            and latitude where the values of the spike pattern want to be retrieved
        

        Examples
        --------
        >>> from seislib.tomography import SeismicTomography
        >>> tomo = SeismicTomography(1, latmin=-20, latmax=20, lonmin=-20, lonmax=20)
        >>> lat = np.mean(tomo.grid.mesh[:, :2], axis=1)
        >>> lon = np.mean(tomo.grid.mesh[:, 2:], axis=1)
        >>> pattern = tomo.spike(ref_value=3, 
        ...                      x0=0, 
        ...                      y0=0, 
        ...                      sigma_x=5, 
        ...                      sigma_y=5,
        ...                      anom_amp=0.2)
        >>> c = pattern(lon, lat)
        >>> img, cb = tomo.plot_map(tomo.grid.mesh, c, projection='Robinson', 
        ...                         show=False)
        >>> cb.set_label('Spike Anomaly')
        """
        
        def gaussian_3d(x, y):
            x_term = (x - x0)**2 / (2 * sigma_x**2)
            y_term = (y - y0)**2 / (2 * sigma_y**2)
            return ref_value + anom_amp * ref_value * np.exp(-(x_term + y_term))
            
        return gaussian_3d
    
    
    def resolution_test(self, mesh, velocity_model, noise=0, ndamp=0, rdamp=0):
        """ Resolution test on an arbitrary input model
        
        This function is called under the hood to perform both the checkerboard
        and the spike test.
        
        Parameters
        ----------
        mesh : ndarray, shape (n, 4)
            Grid cells in seislib format, consisting of n pixels described by 
            lat1, lat2, lon1, lon2 (in degrees). (-180<lon<180, -90<lat<90)
            
        velocity_model : ndarray, shape (n,)
            Synthetic velocity model associated with `mesh`
            
        noise : float, optional
            If > 0, random noise expressed as percentage of the synthetic data
            
        ndamp : float
            Extent of norm damping (default is zero)
            
        rdamp : float
            Extent of roughness damping (default is zero)
            
            
        Returns
        -------
        dict
            The dictionary is structured as follows:

            - 'synth_data': ndarray of shape (m,), where m corresponds to the
              number of rows in the A matrix (i.e., in the data kernel)
                    
            - 'synth_model': inverse of `velocity_model`, used to create the
              synthetic data
                    
            - 'retrieved_model': model retrieved from the inversion of the
              synthetic data
                
            - 'mesh': same as the input `mesh`
        """
        mesh_latmax = np.radians(np.max(mesh[:, 1]))
        mesh_lonmax = np.radians(np.max(mesh[:, 3]))
        data_coords = np.radians(self.data_coords)
        A_test = _compile_coefficients(data_coords=data_coords,
                                       mesh=np.radians(mesh), 
                                       mesh_latmax=mesh_latmax,
                                       mesh_lonmax=mesh_lonmax)
        dist = np.asarray(self.gc_distance(*self.data_coords.T))
        slowness = 1 / velocity_model
        synth_vel = dist / np.dot(A_test*dist.reshape(-1, 1), slowness)
        if noise > 0:
            coeffs = synth_vel * noise
            synth_vel += np.random.uniform(-coeffs, coeffs)         
        retrieved_model = self.solve(A=self.A, 
                                     slowness=1/synth_vel,
                                     mesh=self.grid.mesh,
                                     ndamp=ndamp, 
                                     rdamp=rdamp, 
                                     refvel=np.mean(velocity_model))
        results = {'synth_data': synth_vel, 
                   'synth_model': slowness, 
                   'retrieved_model': retrieved_model, 
                   'mesh': mesh}
        return results
        

    def checkerboard_test(self, kx, ky, regular_grid=False, latmin=None, 
                          latmax=None, lonmin=None, lonmax=None, cell_size=5, 
                          anom_amp=0.1, refvel=None, noise=0, ndamp=0, rdamp=0):
        """ 
        Resolution test, known as "checkerboard test". The method first builds
        synthetic data (velocities) using a :meth:`checkerboard` pattern as 
        input (velocity) model.
        
        Parameters
        ----------
        kx, ky : int , float
            Frequency of anomalies along longitude (`kx`) and latitude (`ky`) within
            the boundaries defined by `lonmin`, `lonmax`, `latmin`, `latmax`
            
        regular_grid : bool
            If False (default), the study area is discretized using an equal-area
            parameterization. Otherwise, a regular grid is employed.
        
        latmin, latmax, lonmin, lonmax : float, optional
            Boundaries of the checkerboard, in degrees. (-180<lon<180, 
            -90<lat<90)
            
        cell_size : int | float
            Size of the checkerboard grid-cell sides (in degrees). Larger values
            are preferrable to avoid the resolution of an overly large inverse 
            problem
            
        anom_amp : float
            Intensity of the anomalies with respect to `ref_value`. A value of 
            0.1 means 10% of `ref_value`
            
        refvel : float, optional
            Original values of the anomalies (black and white squares of a 
            checkerboard). By default, their value is +-1. Therefore, if 
            `anom_extent` is 0.1 (say) their final values will be +-0.1
            
        noise : float, optional
            If > 0, random noise expressed as percentage of the synthetic data
            
        ndamp : float
            Norm damping coefficients (default is zero)
            
        rdamp : float
            Roughness damping coefficients (default is zero)
            
        
        Returns
        -------
        dict
            The dictionary is structured as follows:

            - 'synth_data': ndarray of shape (m,), where m corresponds to the
              number of rows in the A matrix (i.e., in the data kernel)
                    
            - 'synth_model': inverse of `velocity_model`, used to create the
              synthetic data
                    
            - 'retrieved_model': model retrieved from the inversion of the
              synthetic data
                
            - 'mesh': same as the input `mesh`
        """
        if refvel is None:
            refvel = self.refvel
            
        pattern = SeismicTomography.checkerboard(ref_value=refvel,
                                                 kx=kx,
                                                 ky=ky,
                                                 latmin=latmin,
                                                 latmax=latmax,
                                                 lonmin=lonmin,
                                                 lonmax=lonmax,
                                                 anom_amp=anom_amp
                                                 )
        grid = EqualAreaGrid if not regular_grid else RegularGrid
        mesh = grid(cell_size=cell_size,
                         latmin=latmin,
                         lonmin=lonmin, 
                         latmax=latmax,
                         lonmax=lonmax,
                         verbose=False).mesh
        lons = (mesh[:,2] + mesh[:,3]) / 2
        lats = (mesh[:,0] + mesh[:,1]) / 2                                        
        velocity = pattern(lons, lats)
        
        return self.resolution_test(mesh=mesh, 
                                    velocity_model=velocity, 
                                    noise=noise, 
                                    ndamp=ndamp, 
                                    rdamp=rdamp)
        
    
    def spike_test(self, x0, y0, sigma_x, sigma_y, regular_grid=False, 
                   latmin=None, latmax=None, lonmin=None, lonmax=None, 
                   cell_size=5, anom_amp=0.1, refvel=None, noise=0, ndamp=0, 
                   rdamp=0):
        """ 
        Resolution test, known as "spike test". The method first builds
        synthetic data (velocities) using a spike pattern as an input (velocity) 
        model. The spike is built using SeismicTomography.spike.
        
        Parameters
        ----------
        x0, y0 : float
            Central longitude and latitude of the anomaly. (-180<lon<180, 
            -90<lat<90)
            
        sigma_x, sigma_y : float (default is 1)
            Standard deviations in the longitude and latitude directions of the
            3-D Gaussian

        regular_grid : bool
            If False (default), the study area is discretized using an equal-area
            parameterization. Otherwise, a regular grid is employed.


        latmin, latmax, lonmin, lonmax : float, optional
            Boundaries of the checkerboard, in degrees
            
        cell_size : int | float
            Size of the checkerboard grid-cell sides (in degrees). Larger values
            are preferrable to avoid the resolution of an overly large inverse 
            problem
            
        anom_amp : float
            Intensity of the anomalies with respect to `ref_value`. A value of 
            0.1 means 10% of `ref_value`
            
        refvel : float, optional
            Original values of the anomalies (black and white squares of a 
            checkerboard). By default, their value is +-1. Therefore, if 
            `anom_extent` is 0.1 (say) their final values will be +-0.1
            
        noise : float, optional
            If > 0, random noise expressed as percentage of the synthetic data
            
        ndamp : float
            Extent of norm damping (default is zero)
            
        rdamp : float
            Extent of roughness damping (default is zero)
            
        
        Returns
        -------
        Dictionary object structured as follows:

            - 'synth_data': ndarray of shape (m,), where m corresponds to the
              number of rows in the A matrix (i.e., in the data kernel)
                
            - 'synth_model': ndarray of shape (n,), synthetic model (slowness) 
              used to create the synthetic data
                
            - 'retrieved_model': model retrieved from the inversion of the
              synthetic data
                
            - 'mesh': ndarray of shape (n, 4), parameterization associated with 
              the synthetic model
        """
        
        if refvel is None:
            refvel = self.refvel
            
        pattern = SeismicTomography.spike(ref_value=refvel, 
                                          x0=x0, 
                                          y0=y0, 
                                          sigma_x=sigma_x, 
                                          sigma_y=sigma_y, 
                                          anom_amp=anom_amp)
        grid = EqualAreaGrid if not regular_grid else RegularGrid
        mesh = grid(cell_size=cell_size,
                         latmin=latmin,
                         lonmin=lonmin, 
                         latmax=latmax,
                         lonmax=lonmax,
                         verbose=False).mesh
        lons = (mesh[:,2] + mesh[:,3]) / 2
        lats = (mesh[:,0] + mesh[:,1]) / 2                                        
        velocity = pattern(lons, lats)
        
        return self.resolution_test(mesh=mesh, 
                                    velocity_model=velocity, 
                                    noise=noise, 
                                    ndamp=ndamp, 
                                    rdamp=rdamp)


    @classmethod
    def plot_map(cls, mesh, c, ax=None, projection='Mercator', map_boundaries=None, 
                 bound_map=True, colorbar=True, show=True, style='colormesh', 
                 add_features=False, resolution='110m', cbar_dict={}, **kwargs):
        """ 
        Utility function to display the lateral variations of some quantity on 
        a equal-area grid
        
        Parameters
        ----------
        mesh : ndarray, shape (n, 4)
            Grid cells in seislib format, consisting of n pixels described by 
            lat1, lat2, lon1, lon2 (in degrees)
            
        c : ndarray, shape (n,)
            Values associated with the grid cells (mesh)
            
        ax : cartopy.mpl.geoaxes.GeoAxesSubplot
            If not None, `c` is plotted on the `GeoAxesSubplot` instance. 
            Otherwise, a new figure and `GeoAxesSubplot` instance is created
            
        projection : str
            Name of the geographic projection used to create the `GeoAxesSubplot`.
            (Visit the `cartopy website 
            <https://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html>`_ 
            for a list of valid projection names.) If ax is not None, `projection` 
            is ignored. Default is 'Mercator'
            
        map_boundaries : list or tuple of floats, shape (4,), optional
            Lonmin, lonmax, latmin, latmax (in degrees) defining the extent of
            the map
            
        bound_map : bool
            If `True`, the map boundaries will be automatically determined.
            Ignored if `map_boundaries` is not None
        
        colorbar : bool
            If `True` (default), a colorbar associated with `c` is displayed on 
            the side of the map
            
        show : bool
            If `True` (default), the map will be showed once generated
            
        style : {'colormesh', 'contourf', 'contour'}
            Possible options are 'colormesh', 'contourf', and 'contour',
            each corresponding to the homonymous method.
            Default is 'colormesh'
            
        add_features : bool
            If `True`, natural Earth features will be added to the `GeoAxesSubplot`.
            Default is `False`. If `ax` is `None`, it is automatically set to `True`
            
        resolution : {'10m', '50m', '110m'}
            Resolution of the Earth features displayed in the figure. Passed to
            `cartopy.feature.NaturalEarthFeature`. Valid arguments are '110m',
            '50m', '10m'. Default is '110m'
            
        cbar_dict : dict
            Keyword arguments passed to `matplotlib.colorbar.ColorbarBase`
            
        **kwargs
            Additional inputs passed to the 'colormesh', 'contourf', and
            'contour' methods of :class:`SeismicTomography`
            
            
        Returns
        -------
        If `show` is `True`
            `None`
        Otherwise
            an instance of `matplotlib.collections.QuadMesh`, together with an 
            instance of `matplotlib.colorbar.Colorbar` (if `colorbar` is `True`)
        """
        return plotting.plot_map(mesh=mesh, 
                                 c=c, 
                                 ax=ax, 
                                 projection=projection, 
                                 map_boundaries=map_boundaries, 
                                 bound_map=bound_map, 
                                 colorbar=colorbar, 
                                 show=show, 
                                 style=style, 
                                 add_features=add_features, 
                                 resolution=resolution,
                                 cbar_dict=cbar_dict,
                                 **kwargs)


    @classmethod
    def colormesh(cls, mesh, c, ax, **kwargs):
        """
        Adaptation of `matplotlib.pyplot.pcolormesh` to the (adaptive) equal-area 
        grid.
        
        Parameters
        ----------
        mesh : ndarray (n, 4)
            Equal area grid, consisting of n pixels described by lat1, lat2, 
            lon1, lon2
            
        c : list of ndarray (n,)
            Values to plot in each grid cell
            
        ax : cartopy.mpl.geoaxes.GeoAxesSubplot
            If not `None`, the receivers are plotted on the `GeoAxesSubplot` instance. 
            Otherwise, a new figure and GeoAxesSubplot instance is created
            
        **kwargs
            Additional inputs passed to `seislib.plotting.colormesh`
        
        
        Returns
        -------
        img : Instance of `matplotlib.collections.QuadMesh`
        """
        return plotting.colormesh(mesh, c, ax=ax, **kwargs)
        
    
    @classmethod
    def contourf(cls, mesh, c, ax, smoothing=None, **kwargs):
        """
        Adaptation of `matplotlib.pyplot.contourf` to the (adaptive) equal-area 
        grid
        
        Parameters
        ----------
        mesh : ndarray (n, 4)
            Equal area grid, consisting of n pixels described by lat1, lat2, 
            lon1, lon2
            
        c : list of ndarray (n,)
            Values to plot in each grid cell
            
        ax : cartopy.mpl.geoaxes.GeoAxesSubplot
            If not `None`, the receivers are plotted on the `GeoAxesSubplot` 
            instance. Otherwise, a new figure and `GeoAxesSubplot` instance is 
            created
             
        smoothing : float, optional
            Value passed to `scipy.ndimage.filters.gaussian_filter` and used to
            obtain a smooth representation of `c` (default is `None`)
            
        **kwargs
            Additional inputs passed to `seislib.plotting.contourf`
        
        Returns
        -------
        img : Instance of `matplotlib.contour.QuadContourSet`
        """
        
        return plotting.contourf(mesh, c, ax=ax, smoothing=smoothing, **kwargs)


    @classmethod     
    def contour(cls, mesh, c, ax, smoothing=None, **kwargs):
        """
        Adaptation of `matplotlib.pyplot.contour` to the (adaptive) equal area 
        grid
        
        Parameters
        ----------
        mesh : ndarray (n, 4)
            Equal area grid, consisting of n pixels described by lat1, lat2, 
            lon1, lon2
            
        c : list of ndarray (n,)
            Values to plot in each grid cell
            
        ax : cartopy.mpl.geoaxes.GeoAxesSubplot
            If not None, the receivers are plotted on the `GeoAxesSubplot` instance. 
            Otherwise, a new figure and `GeoAxesSubplot` instance is created
            
        smoothing : float, optional
            Value passed to `scipy.ndimage.filters.gaussian_filter` and used to
            obtain a smooth representation of `c` (default is `None`)
            
        **kwargs
            Additional inputs passed to `seislib.plotting.contourf`

        
        Returns
        -------
        img : Instance of `matplotlib.contour.QuadContourSet`
        """
        
        return plotting.contour(mesh, c, ax, smoothing=smoothing, **kwargs)
    
    
    def plot_rays(self, ax=None, show=True, stations_color='r', 
                  paths_color='k', oceans_color='water', lands_color='land', 
                  edgecolor='k', stations_alpha=None, paths_alpha=0.3, 
                  projection='Mercator', resolution='110m', map_boundaries=None, 
                  bound_map=True, paths_width=0.2, **kwargs):
        """ 
        Utility function to display the great-circle paths associated with pairs
        of data coordinates
        
        Parameters
        ----------            
        ax : cartopy.mpl.geoaxes.GeoAxesSubplot
            If not `None`, `c` is plotted on the `GeoAxesSubplot` instance.
            Otherwise, a new figure and `GeoAxesSubplot` instance is created
            
        show : bool
            If `True` (default), the map will be showed once generated. Otherwise
            a `GeoAxesSubplot` instance is returned
            
        stations_color, paths_color : str
            Color of the receivers and of the great-circle paths (see matplotlib 
            documentation for valid color names). Defaults are 'r' (red) and
            'k' (black)
                    
        oceans_color, lands_color : str
            Color of oceans and lands. The arguments are ignored if ax is not
            None. Otherwise, they are passed to `cartopy.feature.NaturalEarthFeature` 
            (to the argument 'facecolor'). Defaults are 'water' and 'land'
    
        edgecolor : str
            Color of the boundaries between, e.g., lakes and land. The argument 
            is ignored if ax is not None. Otherwise, it is passed to 
            cartopy.feature.NaturalEarthFeature (to the argument 'edgecolor'). 
            Default is 'k' (black)
    
        stations_alpha, paths_alpha : float, optional
            Transparency of the stations and of the great-circle paths. Defaults
            are `None` and 0.3
            
        paths_width : float
            Linewidth of the great-circle paths
                    
        projection : str
            Name of the geographic projection used to create the `GeoAxesSubplot`.
            (Visit the `cartopy website 
            <https://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html>`_ 
            for a list of valid projection names.) If ax is not None, `projection` 
            is ignored. Default is 'Mercator'
     
        resolution : {'10m', '50m', '110m'}
            Resolution of the Earth features displayed in the figure. Passed to
            `cartopy.feature.NaturalEarthFeature`. Valid arguments are '110m',
            '50m', '10m'. Default is '110m'. Ignored if `ax` is not `None`.
        
        map_boundaries : list or tuple of floats, shape (4,), optional
            Lonmin, lonmax, latmin, latmax (in degrees) defining the extent of
            the map
            
        bound_map : bool
            If `True`, the map boundaries will be automatically determined.
            Ignored if map_boundaries is not `None`
                        
    
        Returns
        -------
        `None` if `show` is `True`. Otherwise a `GeoAxesSubplot` instance
        """
      
        return plotting.plot_rays(self.data_coords, 
                                  ax=ax, 
                                  show=show, 
                                  stations_color=stations_color, 
                                  paths_color=paths_color, 
                                  oceans_color=oceans_color, 
                                  lands_color=lands_color, 
                                  edgecolor=edgecolor, 
                                  stations_alpha=stations_alpha, 
                                  paths_alpha=paths_alpha, 
                                  projection=projection, 
                                  resolution=resolution,
                                  map_boundaries=map_boundaries, 
                                  bound_map=bound_map, 
                                  paths_width=paths_width, 
                                  **kwargs)


    def plot_colored_rays(self, ax=None, show=True, cmap=scm.roma, vmin=None, 
                          vmax=None, stations_color='k', oceans_color='lightgrey', 
                          lands_color='w', edgecolor='k', stations_alpha=None, 
                          paths_alpha=None, projection='Mercator', resolution='110m',
                          map_boundaries=None, bound_map=True, paths_width=1, 
                          colorbar=True, cbar_dict={}, **kwargs):
        """ 
        Utility function to display the great-circle paths associated with pairs
        of data coordinates, colored according to their respective measurements
        
        Parameters
        ----------      
        ax : cartopy.mpl.geoaxes.GeoAxesSubplot
            If not None, `c` is plotted on the GeoAxesSubplot instance.
            Otherwise, a new figure and GeoAxesSubplot instance is created
            
        show : bool
            If True (default), the map will be showed once generated. Otherwise
            a GeoAxesSubplot instance is returned
            
        cmap : str or Colormap
            If str, it should be a valid matplotlib.cm.colormap name (see 
            matplotlib documentation). Otherwise, it should be the name of a
            colormap available in seislib.colormaps (see also the documentation at 
            https://www.fabiocrameri.ch/colourmaps/)
            
        vmin, vmax : float
            Boundaries of the colormap. If None, the minimum and maximum values
            of `c` will be taken
        
        stations_color : str
            Color of the receivers and of the great-circle paths (see matplotlib 
            documentation for valid color names). Defaults are 'r' (red) and
            'k' (black)
                    
        oceans_color, lands_color : str
            Color of oceans and lands. The arguments are ignored if ax is not
            None. Otherwise, they are passed to `cartopy.feature.NaturalEarthFeature` 
            (to the argument 'facecolor'). Defaults are 'water' and 'land'
    
        edgecolor : str
            Color of the boundaries between, e.g., lakes and land. The argument 
            is ignored if ax is not None. Otherwise, it is passed to 
            cartopy.feature.NaturalEarthFeature (to the argument 'edgecolor'). 
            Default is 'k' (black)
    
        stations_alpha, paths_alpha : float, optional
            Transparency of the stations and of the great-circle paths. Defaults
            are `None` and 0.3
            
        paths_width : float
            Linewidth of the great-circle paths
                    
        projection : str
            Name of the geographic projection used to create the `GeoAxesSubplot`.
            (Visit the `cartopy website 
            <https://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html>`_ 
            for a list of valid projection names.) If ax is not None, `projection` 
            is ignored. Default is 'Mercator'
            
        resolution : {'10m', '50m', '110m'}
            Resolution of the Earth features displayed in the figure. Passed to
            `cartopy.feature.NaturalEarthFeature`. Valid arguments are '110m',
            '50m', '10m'. Default is '110m'. Ignored if `ax` is not `None`.
        
        map_boundaries : list or tuple of floats, shape (4,), optional
            Lonmin, lonmax, latmin, latmax (in degrees) defining the extent of
            the map
            
        bound_map : bool
            If `True`, the map boundaries will be automatically determined.
            Ignored if `map_boundaries` is not `None`
        
        colorbar : bool
            If `True` (default), a colorbar associated with `c` is displayed on 
            the side of the map        
        
        cbar_dict : dict
            Keyword arguments passed to `matplotlib.colorbar.ColorbarBase`
        
        **kwargs
            Additional keyword arguments passed to ax.plot
        
        Returns
        -------
        If `show` is `True`
            `None` 
        Otherwise 
            `GeoAxesSubplot` instance together with an instance of 
            `matplotlib.colorbar.Colorbar` (if `colorbar` is `True`)
        """ 
        return plotting.plot_colored_rays(data_coords=self.data_coords, 
                                          c=self.velocity,
                                          ax=ax, 
                                          show=show, 
                                          cmap=cmap, 
                                          vmin=vmin, 
                                          vmax=vmax, 
                                          stations_color=stations_color, 
                                          oceans_color=oceans_color, 
                                          lands_color=lands_color, 
                                          edgecolor=edgecolor, 
                                          stations_alpha=stations_alpha, 
                                          paths_alpha=paths_alpha, 
                                          projection=projection, 
                                          resolution=resolution,
                                          map_boundaries=map_boundaries, 
                                          bound_map=bound_map, 
                                          paths_width=paths_width, 
                                          colorbar=colorbar, 
                                          cbar_dict=cbar_dict, 
                                          **kwargs)


        
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Fabrizio Magrini
@email1: fabrizio.magrini@uniroma3.it
@email2: fabrizio.magrini90@gmail.com
"""



class DispersionCurveException(Exception):
    """
    Exception raised when a dispersion curve could not be extracted from the 
    data
    """

    def __init__(self):
        self.message = 'It was not possible to retrieve a dispersion curve'
        super().__init__(self.message)

    def __str__(self):
        return self.message


class TimeSpanException(Exception):
    """
    Exception raised when no common time span is found in two obspy traces or 
    streams.
    """

    def __init__(self, *args, message=None):
        if message is not None:
            self.message = message
        else:
            self.message = 'No common time span found.'
        for arg in args:
            self.message += '\n%s'%arg
        super().__init__(self.message)

    def __str__(self):
        return self.message


class NonFiniteDataException(Exception):
    """
    Exception raised when the data should be strictly finite but contain
    infinite or nan values.
    """

    def __init__(self, *args):
        self.message = 'The data should be strictly finite, but contain either'
        self.message += ' infinite or nan values.'
        for arg in args:
            self.message += '\n%s'%arg
        super().__init__(self.message)

    def __str__(self):
        return self.message






__major_version__  = 0
__minor_version__  = 6
__patch__          = 8
__version_tuple__ = (
    __major_version__,
    __minor_version__,
    __patch__
)
__version_number__ = ".".join([str(v) for v in __version_tuple__])
__version__ = f"{__version_number__}"



"""
===========================================
Utility Functions (:mod:`seislib.utils`)
===========================================

This module provides support for several useful functions for
processing and analysing seismic data. These mainly build on
`ObsPy <https://docs.obspy.org/>`_, and include:

- Vectorized calculation of great-circle distances, azimuths, and back
  azimuths
  
- Slicing / resampling of pairs of seismograms so as to adapt them to 
  a common time window / sampling rate
  
- Pre-processing operations such as zero-padding and bandpass
  filtering
  
- Rotation of horizontal-component recordings to an arbitrary amount
  of degrees. SeisLib's implementation also takes into account not 
  uncommon ValueErrors raised by ObsPy due to the presence of slightly 
  (sub-sample) differences in the time spanned by the two traces 

- Interpolation of scattered data on equal-area grids

- Calculation of the Pearson correlation coefficient between 
  physical parameters defined on equal-area grids
  
"""
from .utils import *

